//========================================================================================
//
//	File:				FWGDev.cpp
//	Release Version:	$ 1.0d11 $
//
//	Copyright:	 1993, 1995 by Apple Computer, Inc., all rights reserved.
//
//========================================================================================

#include "FWOS.hpp"

#if defined(FW_PROFILER_CALLS) && defined(__MWERKS__)
#pragma profile on
#endif

#ifndef FWGDEV_H
#include "FWGDev.h"
#endif

#ifndef FWREGION_H
#include "FWRegion.h"
#endif

#ifndef FWPAT_H
#include "FWPat.h"
#endif

#ifndef FWGRGLOB_H
#include "FWGrGlob.h"
#endif

#ifndef FWINK_H
#include "FWInk.h"
#endif

#ifndef FWSTYLE_H
#include "FWStyle.h"
#endif

#ifndef FWFONT_H
#include "FWFont.h"
#endif

#ifndef FWGC_H
#include "FWGC.h"
#endif

#ifndef FWGRUTIL_H
#include "FWGrUtil.h"
#endif

#ifndef FWODGEOM_H
#include "FWODGeom.h"
#endif

// ----- Platform Includes -----

#if defined(FW_BUILD_MAC) && !defined(__LOWMEM__)
#include <LowMem.h>
#endif

#if defined(FW_BUILD_MAC) && !defined(__PRINTING_)
#include <Printing.h>
#endif

#ifdef FW_BUILD_MAC
#include <stdio.h>	// Need sprintf for PostScript clipping
#endif

// ----- OpenDoc Includes -----

#ifndef SOM_ODCanvas_xh
#include <Canvas.xh>
#endif

#ifndef SOM_ODTransform_xh
#include <Trnsform.xh>
#endif

// ----- Standard C Includes -----

#include <math.h>

//========================================================================================
//	RunTime Info
//========================================================================================

#if FW_LIB_EXPORT_PRAGMAS
#pragma lib_export on
#endif

#pragma segment FWGraphics_Device

//========================================================================================
// Macros
//========================================================================================

#ifdef FW_BUILD_WIN

#define FW_CHECK_PLATFORM_CANVAS \
	FW_ASSERT(fPlatformCanvas != NULL);

#endif

#ifdef FW_BUILD_MAC

#define FW_CHECK_PLATFORM_CANVAS \
	FW_ASSERT(fPlatformCanvas != NULL); \
	FW_ASSERT(fPlatformCanvas == FW_QDGlobals.thePort);

#endif

//========================================================================================
//	class FW_CGraphicDevice
//========================================================================================

//----------------------------------------------------------------------------------------
//	FW_CGraphicDevice::FW_CGraphicDevice
//----------------------------------------------------------------------------------------

FW_CGraphicDevice::FW_CGraphicDevice(Environment *ev, ODCanvas* odCanvas) :
	FW_MRefCount(),
	fODCanvas(odCanvas),
	fPlatformCanvas(NULL),
#ifdef FW_BUILD_MAC
	fPattern(FW_kBlackPat),
#endif
	fOpenDeviceCount(0),
	fSuspended(FALSE),
	fContext(NULL),
	fResolution(FW_kZeroPoint),
	fEV(ev)
{
	FW_ASSERT(fODCanvas != NULL);
}

//----------------------------------------------------------------------------------------
//	FW_CGraphicDevice::FW_CGraphicDevice
//----------------------------------------------------------------------------------------

FW_CGraphicDevice::FW_CGraphicDevice(Environment *ev, ODPlatformCanvas platformCanvas) :
	FW_MRefCount(),
	fODCanvas(NULL),
	fPlatformCanvas(platformCanvas),
#ifdef FW_BUILD_MAC
	fPattern(FW_kBlackPat),
#endif
	fOpenDeviceCount(0),
	fSuspended(FALSE),
	fContext(NULL),
	fResolution(FW_kZeroPoint),
	fEV(ev)
{
	FW_ASSERT(fPlatformCanvas != NULL);
}

//---------------------------------------------------------------------------------------
// 	FW_CGraphicDevice::~FW_CGraphicDevice
//---------------------------------------------------------------------------------------

FW_CGraphicDevice::~FW_CGraphicDevice()
{
	FW_ASSERT(fOpenDeviceCount == 0);
	FW_ASSERT(!fSuspended);
}

//----------------------------------------------------------------------------------------
//	FW_CGraphicDevice::GetResolution
//----------------------------------------------------------------------------------------
FW_CPoint FW_CGraphicDevice::GetResolution(Environment *ev) const
{
	if (fResolution == FW_kZeroPoint)
	{
		FW_ASSERT(fPlatformCanvas != NULL);
		
		FW_CGraphicDevice* self = (FW_CGraphicDevice *) this;

#ifdef FW_BUILD_WIN
		self->fResolution.x = FW_IntToFixed(::GetDeviceCaps(fPlatformCanvas, LOGPIXELSX));
		self->fResolution.y = FW_IntToFixed(::GetDeviceCaps(fPlatformCanvas, LOGPIXELSY));
#endif
#ifdef FW_BUILD_MAC
		self->fResolution.x = FW_IntToFixed(72);
		self->fResolution.y = FW_IntToFixed(72);
#endif
	}

	return fResolution;
}

//----------------------------------------------------------------------------------------
//	FW_CGraphicDevice::SetResolution
//----------------------------------------------------------------------------------------

void FW_CGraphicDevice::SetResolution(const FW_CPoint& resolution)
{
	fResolution = resolution;
}

#ifdef FW_BUILD_WIN
//----------------------------------------------------------------------------------------
//	FW_CGraphicDevice::IsMetaFileCanvas
//----------------------------------------------------------------------------------------

FW_Boolean FW_CGraphicDevice::IsMetaFileCanvas() const
{
	FW_ASSERT(fPlatformCanvas);
	return ::GetDeviceCaps(fPlatformCanvas, TECHNOLOGY) == DT_METAFILE;
}
#endif

//---------------------------------------------------------------------------------------
//	FW_CGraphicDevice::OpenDevice
//---------------------------------------------------------------------------------------

FW_SDeviceState* FW_CGraphicDevice::OpenDevice(Environment *ev, FW_CGraphicContext* context)
{
// [KVV]	FW_ASSERT(!fSuspended);

	FW_ASSERT(fContext == NULL);
	fContext = context;

	// ----- Get the platform canvas -----
	if (fPlatformCanvas == NULL)
	{
#ifdef FW_BUILD_WIN
		fPlatformCanvas = fODCanvas->GetPlatformCanvas(ev, kODWindows);	
#endif
#ifdef FW_BUILD_MAC
		fPlatformCanvas = fODCanvas->GetPlatformCanvas(ev, kODQuickDraw);	
#endif
	}		
	FW_ASSERT(fPlatformCanvas != NULL);

	// ----- Save first the current state -----
	FW_SDeviceState* deviceState = GetState();		// GetState set the port

	// ----- Reset Settings -----
	if (fOpenDeviceCount == 0)
	{
		// Update the resolution
		FW_CPoint res = GetResolution(ev);
		
#ifdef FW_BUILD_WIN
		fTextColor = ::GetTextColor(fPlatformCanvas);
		fBkColor = ::GetBkColor(fPlatformCanvas);
		fPenSize.Set(0, 0);
		
		::SetMapMode(fPlatformCanvas, MM_TEXT);
#endif
#ifdef FW_BUILD_MAC
		ResetSettings();
		fMacPostScriptClip = FALSE;
#endif
	}
	
	fSuspended = FALSE;		// we are opening it so it is not suspended
	fOpenDeviceCount++;

	// ----- Set the origin offset -----
	UpdateOriginForContext();

	return deviceState;
}

//---------------------------------------------------------------------------------------
//	FW_CGraphicDevice::CloseDevice
//---------------------------------------------------------------------------------------

void FW_CGraphicDevice::CloseDevice(Environment *ev, FW_SDeviceState* deviceState)
{
	FW_CHECK_PLATFORM_CANVAS

	FW_ASSERT(!fSuspended);
	FW_ASSERT(fOpenDeviceCount != 0);

#ifdef FW_BUILD_MAC
// ----- [HLX] OpenDoc Bug 
	// I should not have to restore the Grafport but
	// ODFacet::DrawActiveBorder doesn't reset the foreground
	// and background color correctly so....
	::PenNormal();
	::RGBForeColor(&FW_kRGBBlack);
	::RGBBackColor(&FW_kRGBWhite);
// ----- [HLX] OpenDoc Bug
#endif

	// ----- Restore origin and clip -----
#ifdef FW_BUILD_WIN
	::SetWindowOrgEx(fPlatformCanvas, deviceState->fWindOrg.x, deviceState->fWindOrg.y, NULL);
	::SetMapMode(fPlatformCanvas, deviceState->fMapMode);
#endif
#ifdef FW_BUILD_MAC
	if (fMacPostScriptClip)
		MacEndPostScriptClip();

	::SetOrigin(deviceState->fOrigin.h, deviceState->fOrigin.v);

	if (FW_MacIsColorPort(FW_QDGlobals.thePort))	// [LW7]
	{
		::SetClip(deviceState->fClip);
	}

	::DisposeRgn(deviceState->fClip);
	deviceState->fClip = NULL;

	// ----- Restore previous port -----
	::SetPort(deviceState->fPreviousPlatformCanvas);
#endif

#ifdef FW_BUILD_MAC
	// ----- Restore the suspended flag
	// [KVV] CloseDevice should set fSuspended to FALSE!
	fSuspended = deviceState->fSuspended;
#endif

	// ----- Decrement fOpenDeviceCount -----
	fOpenDeviceCount--;
	
	// ----- If fOpenDeviceCount == 0 the device is not used anymore.
	if (fOpenDeviceCount == 0)
	{	
#ifdef FW_BUILD_WIN
		// ----- Unselect our object -----
		fGDIPen.UnselectObject(fPlatformCanvas);
		fGDIBrush.UnselectObject(fPlatformCanvas);
		fGDIFont.UnselectObject(fPlatformCanvas);
		
		// ----- Reset also textColor, bkColor etc to their default value -----
		::SetTextColor(fPlatformCanvas, FW_kRGBBlack);
		::SetBkColor(fPlatformCanvas, FW_kRGBWhite);
		::SetROP2(fPlatformCanvas, R2_COPYPEN);
		::SetBkMode(fPlatformCanvas, OPAQUE);
#endif

		if (fODCanvas != NULL)
		{
#ifdef FW_BUILD_WIN
			fODCanvas->ReleasePlatformCanvas(ev);
#endif
		}
		
		fPlatformCanvas = NULL;
	}

	// ----- We don't need the deviceState anymore -----
	delete deviceState;

	// ----- Not valid anymore -----
	fContext = NULL;
}

//---------------------------------------------------------------------------------------
//	FW_CGraphicDevice::GetState
//---------------------------------------------------------------------------------------

FW_SDeviceState* FW_CGraphicDevice::GetState()
{
	FW_ASSERT(fPlatformCanvas != NULL);
	FW_SDeviceState* deviceState = new FW_SDeviceState();
	
#ifdef FW_BUILD_WIN
	deviceState->fMapMode = ::GetMapMode(fPlatformCanvas);
	::GetWindowOrgEx(fPlatformCanvas, &deviceState->fWindOrg);
#endif
#ifdef FW_BUILD_MAC
	// ----- Save previous grafport -----
	::GetPort(&deviceState->fPreviousPlatformCanvas);
	FW_ASSERT(deviceState->fPreviousPlatformCanvas != NULL);
	
	// ----- Set the new grafport -----
	::SetPort(fPlatformCanvas);
	
	// ----- Save the origin and the clip of our port 	
	deviceState->fOrigin.h = fPlatformCanvas->portRect.left;
	deviceState->fOrigin.v = fPlatformCanvas->portRect.top;

	deviceState->fClip = ::FW_CopyRegion(fPlatformCanvas->clipRgn);

	deviceState->fSuspended = fSuspended;
#endif

	return deviceState;
}

//---------------------------------------------------------------------------------------
//	FW_CGraphicDevice::Suspend
//---------------------------------------------------------------------------------------

FW_SSuspendResumeState* FW_CGraphicDevice::Suspend()
{
// [KVV]	FW_ASSERT(!fSuspended);
	FW_ASSERT(fOpenDeviceCount != 0);	// Should not have to Suspend if not opened

//	[KVV] When called from the drag manager, the current port has already been changed
//	FW_CHECK_PLATFORM_CANVAS

#ifdef FW_BUILD_WIN
	FW_SSuspendResumeState* deviceState = new FW_SSuspendResumeState();
	deviceState->fSavedDC = ::SaveDC(fPlatformCanvas);
#endif
#ifdef FW_BUILD_MAC
	FW_SSuspendResumeState* deviceState = GetState();
#endif

	deviceState->fSavedContext = fContext;
	fContext = NULL;

	fSuspended = TRUE;

	return deviceState;
}

//---------------------------------------------------------------------------------------
//	FW_CGraphicDevice::Resume
//---------------------------------------------------------------------------------------

void FW_CGraphicDevice::Resume(FW_SSuspendResumeState* deviceState)
{	
// [KVV]	FW_ASSERT(fSuspended);
	FW_ASSERT(fPlatformCanvas != NULL);
	
	FW_ASSERT(fContext == NULL);
	fContext = deviceState->fSavedContext;

#ifdef FW_BUILD_WIN
	::RestoreDC(fPlatformCanvas, deviceState->fSavedDC);
#endif
#ifdef FW_BUILD_MAC
//	[KVV] Drag and drop (also see comment in Suspend)
//	FW_ASSERT(deviceState->fPreviousPlatformCanvas == fPlatformCanvas);

	::SetPort(fPlatformCanvas);	// The port may have changed
	
	::SetOrigin(deviceState->fOrigin.h, deviceState->fOrigin.v);
	::SetClip(deviceState->fClip);
	::DisposeRgn(deviceState->fClip);
	deviceState->fClip = NULL;

	fSuspended = deviceState->fSuspended;		// [KVV] Resume should "fSuspended = FALSE"

	// ----- We need to reset our settings
	ResetSettings();		
#endif
	
	// ----- We don't need the suspendResumeState anymore -----
	delete deviceState;
	
	// ----- We are not suspended anymore -----
	fSuspended = FALSE;
}

//---------------------------------------------------------------------------------------
//	FW_CGraphicDevice::CanvasChanged
//---------------------------------------------------------------------------------------

void FW_CGraphicDevice::CanvasChanged(Environment* ev, ODCanvas* newCanvas)
{
	FW_ASSERT(!fSuspended);
	FW_ASSERT(fOpenDeviceCount == 0);
	FW_ASSERT(fODCanvas != NULL);	// Should not change the canvas if didn't have a canvas already
	
#ifdef FW_BUILD_WIN
	if(fPlatformCanvas != NULL)
	{
		fODCanvas->ReleasePlatformCanvas(ev);
		fPlatformCanvas = NULL;
	}
#endif
	
	fODCanvas = newCanvas;
	
#ifdef FW_BUILD_WIN
	fPlatformCanvas = fODCanvas->GetPlatformCanvas(ev, kODWindows);
#endif
#ifdef FW_BUILD_MAC
	fPlatformCanvas = fODCanvas->GetPlatformCanvas(ev, kODQuickDraw);
#endif
}

//---------------------------------------------------------------------------------------
//	FW_CGraphicDevice::MappingChanged
//---------------------------------------------------------------------------------------

void FW_CGraphicDevice::MappingChanged()
{
	// ----- Set the origin offset -----
	UpdateOriginForContext();
}

//---------------------------------------------------------------------------------------
//	FW_CGraphicDevice::UpdateOriginForContext
//---------------------------------------------------------------------------------------

void FW_CGraphicDevice::UpdateOriginForContext()
{
	FW_SPlatformPoint offsetOrigin = fContext->GetOriginOffset();

	offsetOrigin.Set(
		-offsetOrigin.X(),
		-offsetOrigin.Y());

	SetOrigin(offsetOrigin);
}

//---------------------------------------------------------------------------------------
//	FW_CGraphicDevice::SetOrigin
//---------------------------------------------------------------------------------------

void FW_CGraphicDevice::SetOrigin(FW_SPlatformPoint origin)
{
	FW_ASSERT(!fSuspended);
	FW_ASSERT(fOpenDeviceCount != 0);

	FW_CHECK_PLATFORM_CANVAS

#ifdef FW_BUILD_WIN
	::SetWindowOrgEx(fPlatformCanvas, origin.x, origin.y, NULL);
#endif
#ifdef FW_BUILD_MAC
	GrafPtr curPort;
	::GetPort(&curPort);

	FW_SPlatformPoint oldOrigin(curPort->portRect.left, curPort->portRect.top);
	if(oldOrigin.h != origin.h || oldOrigin.v != origin.v)
	{
		// Set the port origin
		::SetOrigin(origin.h, origin.v);

		// Shift the clip region
		ODRgnHandle rgn = GetClip();
		::OffsetRgn(rgn, origin.h - oldOrigin.h, origin.v - oldOrigin.v);
		SetClip(rgn);
		::DisposeRgn(rgn);
	}
#endif
}

//---------------------------------------------------------------------------------------
//	FW_CGraphicDevice::SetClip
//---------------------------------------------------------------------------------------

void FW_CGraphicDevice::SetClip(ODRgnHandle clipRegion)
{
	FW_ASSERT(!fSuspended);
	FW_ASSERT(fOpenDeviceCount != 0);

	FW_CHECK_PLATFORM_CANVAS

#ifdef FW_BUILD_WIN
	::SelectClipRgn(fPlatformCanvas, clipRegion);
#endif
#ifdef FW_BUILD_MAC
	if (fMacPostScriptClip)
		MacEndPostScriptClip();
	
	if (FW_MacIsColorPort(FW_QDGlobals.thePort))	// [LW7]
	{
		::SetClip(clipRegion);		// SetClip copies the region
	}

	if (fODCanvas != NULL && fODCanvas->HasPlatformPrintJob(fEV, kODQuickDraw))
	{
#define bDevLaser			3	// copied from "PrPrivate.a"

		THPrint thPrint = (THPrint) fODCanvas->GetPlatformPrintJob(fEV, kODQuickDraw);
		FW_Boolean bPostScript = (((**thPrint).prStl.wDev >> 8) == bDevLaser);
		
		if (bPostScript && (*clipRegion)->rgnSize != 10)
		{
			// If the region is not rectangular
			MacBeginPostScriptClip(clipRegion);
		}
	}
#endif
}

//---------------------------------------------------------------------------------------
//	FW_CGraphicDevice::GetClip
//---------------------------------------------------------------------------------------

ODRgnHandle FW_CGraphicDevice::GetClip() const
{
	FW_ASSERT(!fSuspended);
	FW_ASSERT(fOpenDeviceCount != 0);

	FW_CHECK_PLATFORM_CANVAS

#ifdef FW_BUILD_WIN
	HRGN clipHandle = ::CreateRectRgn(0, 0, 0, 0);
	::GetClipRgn(fPlatformCanvas, clipHandle);
#endif
#ifdef FW_BUILD_MAC
	ODRgnHandle clipHandle = ::NewRgn();
	::GetClip(clipHandle);
#endif

	return clipHandle;
}

//---------------------------------------------------------------------------------------
//	FW_CGraphicDevice::SetClipRect
//---------------------------------------------------------------------------------------

void FW_CGraphicDevice::SetClipRect(const FW_SPlatformRect& clipRect)
{
	FW_ASSERT(!fSuspended);
	FW_ASSERT(fOpenDeviceCount != 0);

	FW_CHECK_PLATFORM_CANVAS

#ifdef FW_BUILD_WIN
	ODRgnHandle clipRegion = ::CreateRectRgnIndirect(&clipRect);
	::SelectClipRgn(fPlatformCanvas, clipRegion);
	::DeleteObject(clipRegion);
#endif
#ifdef FW_BUILD_MAC
	if (fMacPostScriptClip)
		MacEndPostScriptClip();

	::ClipRect(&clipRect);
#endif
}

//---------------------------------------------------------------------------------------
//	FW_CGraphicDevice::IntersectClipRect
//---------------------------------------------------------------------------------------

void FW_CGraphicDevice::IntersectClipRect(const FW_SPlatformRect& clipRect)
{
	FW_ASSERT(!fSuspended);
	FW_ASSERT(fOpenDeviceCount != 0);

	FW_CHECK_PLATFORM_CANVAS

#ifdef FW_BUILD_WIN
	::IntersectClipRect(fPlatformCanvas,
						clipRect.left, clipRect.top, 
						clipRect.right, clipRect.bottom);
#endif
#ifdef FW_BUILD_MAC
	if (fMacPostScriptClip)
		MacEndPostScriptClip();

	ODRgnHandle rgn = ::NewRgn();
	::RectRgn(rgn, &clipRect);
	ODRgnHandle curClip = GetClip();
	::SectRgn(curClip, rgn, curClip);
	::DisposeRgn(rgn);
	SetClip(curClip);
	::DisposeRgn(curClip);
#endif
}

//---------------------------------------------------------------------------------------
//	FW_CGraphicDevice::GetClipRect
//---------------------------------------------------------------------------------------

void FW_CGraphicDevice::GetClipRect(FW_SPlatformRect& clipRect) const
{
	FW_ASSERT(!fSuspended);
	FW_ASSERT(fOpenDeviceCount != 0);

	FW_CHECK_PLATFORM_CANVAS

#ifdef FW_BUILD_WIN
	::GetClipBox(fPlatformCanvas, &clipRect);
#endif
#ifdef FW_BUILD_MAC
	clipRect = (*FW_QDGlobals.thePort->clipRgn)->rgnBBox;
#endif
}

#ifdef FW_BUILD_MAC
//----------------------------------------------------------------------------------------
//	FW_CGraphicDevice::SetHiliteMode
//----------------------------------------------------------------------------------------

void FW_CGraphicDevice::SetHiliteMode()
{
	unsigned char HiliteModeValue = LMGetHiliteMode();
	::BitClr(&HiliteModeValue, pHiliteBit);
	LMSetHiliteMode(HiliteModeValue);
}
#endif

#ifdef FW_BUILD_MAC
//----------------------------------------------------------------------------------------
//	FW_CGraphicDevice::SetInGrafPort
//----------------------------------------------------------------------------------------
//	[HLX] Maybe it is as fast to just set than testing and setting???

void FW_CGraphicDevice::SetInGrafPort()
{
	FW_ASSERT(!fSuspended);
	FW_ASSERT((fOpenDeviceCount != 0) && (fPlatformCanvas == FW_QDGlobals.thePort));
	
	if (fChangeFlag & FW_kForeColorChanged)
		::RGBForeColor(&fForeColor);
		
	if (fChangeFlag & FW_kBackColorChanged)
		::RGBBackColor(&fBackColor);
		
	if (fChangeFlag & FW_kHiliteColorChanged)
		::HiliteColor(&fHiliteColor);
		
	if (fChangeFlag & FW_kPatternChanged)
		fPattern->MacSetInCurPort();
		
	if (fPlatformCanvas->pnSize.h != fHPenSize || fPlatformCanvas->pnSize.v != fVPenSize)
		::PenSize(fHPenSize, fVPenSize);
		
	if (fPlatformCanvas->txFont != fFontID)
		::TextFont(fFontID);
		
	if (fPlatformCanvas->txSize != fFontSize)
		::TextSize(fFontSize);
		
	if (fPlatformCanvas->txFace != fFontStyle)
		::TextFace(fFontStyle);
		
	if (fPlatformCanvas->pnMode != fPenMode)
		::PenMode(fPenMode);
		
	if (fPlatformCanvas->txMode != fTextMode)
		::TextMode(fTextMode);
	
	fChangeFlag = 0;
}
#endif

#ifdef FW_BUILD_MAC
//----------------------------------------------------------------------------------------
//	FW_CGraphicDevice::ResetSettings
//----------------------------------------------------------------------------------------

void FW_CGraphicDevice::ResetSettings()
{
	FW_ASSERT((fPlatformCanvas == FW_QDGlobals.thePort));

	::GetForeColor(&fForeColor);
	::GetBackColor(&fBackColor);
	
	LMGetHiliteRGB(&fHiliteColor);		// [HLX] I am not sure about that
	::HiliteColor(&fHiliteColor);
	
	fPattern = FW_kBlackPat;
	fPattern->MacSetInCurPort();
	
	fHPenSize = fPlatformCanvas->pnSize.h;
	fVPenSize = fPlatformCanvas->pnSize.v;
	fFontID = fPlatformCanvas->txFont;
	fFontSize = fPlatformCanvas->txSize;
	fFontStyle = fPlatformCanvas->txFace;
	fPenMode = fPlatformCanvas->pnMode;
	fTextMode = fPlatformCanvas->txMode;
	
	fChangeFlag = 0;
}
#endif

#ifdef FW_BUILD_MAC
//----------------------------------------------------------------------------------------
//	GetMacTransferMode
//----------------------------------------------------------------------------------------

static short GetMacTransferMode(FW_TransferModes transferMode)
{
	if ((transferMode & 0xFFFF0000) == 0)
		return (short)transferMode;		// Native transfer mode
	else
	{
		FW_ASSERT(transferMode <= FW_LastTransferMode);
		return FW_gTransferModes[transferMode & 0x0000FFFF];
	}
}
#endif

#ifdef FW_BUILD_MAC
//----------------------------------------------------------------------------------------
//	FW_CGraphicDevice::SelectInk
//----------------------------------------------------------------------------------------

void FW_CGraphicDevice::SelectInk(const FW_PInk& ink, 
								FW_EShapeCategories shapeCategory, 
								FW_ERenderVerbs renderVerb)
{
	FW_ASSERT((const void*)ink != NULL);

	FW_CColor foreColor, backColor;
	FW_TransferModes transferMode = ink->GetTransferMode();
	
	ink->GetForeColor(foreColor);
	ink->GetBackColor(backColor);
	
	if (transferMode == FW_kHilite)
		SetHiliteColor(foreColor);

	switch (shapeCategory)
	{
		case FW_kLineShape:
			if (transferMode == FW_kErase)
			{
				SetForeColor(backColor);
				SetBackColor(backColor);
				SetPenMode(GetMacTransferMode(transferMode));
			}
			else if (transferMode == FW_kInvert)
			{
				SetForeColor(foreColor);
				SetBackColor(backColor);
				SetPenMode(patXor);
			}
			else
			{
				SetForeColor(foreColor);
				SetBackColor(backColor);
				SetPenMode(GetMacTransferMode(transferMode));
			}
			break;
			
		case FW_kGeometricShape:
			if (renderVerb == FW_kFrame)
			{
				if (transferMode == FW_kErase)
				{
					SetForeColor(backColor);
					SetBackColor(backColor);
					SetPenMode(patCopy);
				}
				else if (transferMode == FW_kInvert)
				{
					SetForeColor(foreColor);
					SetBackColor(backColor);
					SetPenMode(patXor);
				}
				else
				{
					SetForeColor(foreColor);
					SetBackColor(backColor);
					SetPenMode(GetMacTransferMode(transferMode));
				}
				break;
			}
			else
			{
				SetForeColor(foreColor);
				SetBackColor(backColor);
				if (transferMode == FW_kInvert || transferMode == FW_kErase)
					SetPenMode(patCopy);	// I will use ::InvertXXX or ::EraseXXX
				else
					SetPenMode(GetMacTransferMode(transferMode));
			}
			
			break;
			
		case FW_kTypographicShape:
			SetBackColor(backColor);
			if (transferMode == FW_kInvert)
			{
				SetForeColor(foreColor);
				SetPenMode(patXor);
			}
			else if (transferMode == FW_kErase)
			{
				SetForeColor(backColor);
				SetPenMode(patCopy);
			}
			else
			{
				SetForeColor(foreColor);
				SetTextMode(GetMacTransferMode(transferMode));
			}
			break;
		
		case FW_kImageShape:
			SetForeColor(foreColor);
			SetBackColor(backColor);
			SetPenMode(GetMacTransferMode(transferMode));
			break;
	}
}
#endif

#ifdef FW_BUILD_MAC
//----------------------------------------------------------------------------------------
//	FW_CGraphicDevice::SelectStyle
//----------------------------------------------------------------------------------------

FW_Boolean FW_CGraphicDevice::SelectStyle(const FW_PStyle& style, 
									FW_TransferModes transferMode)
{
	FW_ASSERT((const void*)style != NULL);

	if (transferMode == FW_kInvert || transferMode == FW_kErase)
		SetPattern(FW_kBlackPat);
	else
		SetPattern(style->GetPattern());

	FW_EStyleDash dash = style->GetDashStyle();
	FW_CFixed penSize = style->GetPenSize();
	
	FW_SPlatformPoint plfmPt;
	if (!penSize)
	{
		plfmPt.Set(1, 1);
	}
	else
	{
		plfmPt = fContext->LogicalToDevice(penSize, penSize);
		
		if (plfmPt.h == 0)
			plfmPt.h = 1;
			
		if (plfmPt.v == 0)
			plfmPt.v = 1;
	}

	SetPenSize(plfmPt.h, plfmPt.v);
	
	return dash != FW_kSolidLine && plfmPt.h <= 1 && plfmPt.v <= 1;
}
#endif

#ifdef FW_BUILD_MAC
//----------------------------------------------------------------------------------------
//	FW_CGraphicDevice::SetPattern
//----------------------------------------------------------------------------------------

void FW_CGraphicDevice::SetPattern(const FW_PPattern& pattern)
{
	fPattern =	pattern;
	fChangeFlag |= FW_kPatternChanged;
}
#endif

#ifdef FW_BUILD_MAC

#define kPostScriptBegin	190		// Picture-comments for PostScript printing
#define kPostScriptEnd		191
#define kTextIsPostScript	194

//----------------------------------------------------------------------------------------
//	FW_CGraphicDevice::MacBeginPostScriptClip
//----------------------------------------------------------------------------------------

void FW_CGraphicDevice::MacBeginPostScriptClip(RgnHandle rgn)
{
	FW_ASSERT(!fMacPostScriptClip);

	//	In order for the PostScript commands we'll be emitting below to sync up
	//	properly with other PS code the driver emits, we have to get something else
	//	to emit right now as a flush. So we'll draw a tiny rectangle and make sure
	//	it falls outside the clipping bounds.
	
	Rect bbox = (*rgn)->rgnBBox;

	Rect r;
	r.right = bbox.left;
	if (r.right == -32767)
		r.right = bbox.right + 1;
	r.bottom = bbox.top;
	if (r.bottom == -32767)
		r.bottom = bbox.bottom + 1;
	r.left = r.right - 1;
	r.top  = r.bottom - 1;
	::PaintRect(&r);

	// Copy a region to a shape so it can be converted to a polygon
	RgnHandle rgnCopy = ::NewRgn();
	::CopyRgn(rgn, rgnCopy);
	ODShape* shape = FW_NewODShape(fEV, rgnCopy);

	// Now emit the polygon as a PostScript path and clip to it
	ODPolygon poly;
	shape->CopyPolygon(fEV, &poly);
	
	FW_ODPolygonData* polyData = (FW_ODPolygonData*) poly._buffer;

	::PicComment(kPostScriptBegin, 0, kODNULL);
	::PicComment(kTextIsPostScript, 0, kODNULL);

	::DrawString("\ppse gsave");			// Cancels out the 'psb' generated by the LW driver
	::DrawString("\pnewpath");

	char buf[128];
	FW_ODContour* cont = polyData->fContours;
	for (long n = polyData->fContourCount; n > 0; -- n, cont = cont->NextContour())
	{
		const FW_CPoint* v = cont->fVertex;
		long m = cont->fVertexCount;
		if (m > 2)
		{
			::DrawText(buf, 0,
				sprintf(buf, "%.2f %.2f moveto", v->x.AsDouble(), v->y.AsDouble()));

			while (--m > 0)
			{
				++ v;
				::DrawText(buf, 0,
					sprintf(buf,"%.2f %.2f lineto", v->x.AsDouble(), v->y.AsDouble()));
			}
		}
	}

	::DrawString("\pclosepath clip");	// Adding newpath would be nice but LW7 no likee

	::DrawString("\ppsb");				// Cancels out the forthcoming 'pse'
	::PicComment(kPostScriptEnd, 0,kODNULL);

	delete[] poly._buffer;
	
	shape->Release(fEV);

	fMacPostScriptClip = TRUE;
}
#endif

#ifdef FW_BUILD_MAC
//----------------------------------------------------------------------------------------
//	FW_CGraphicDevice::MacEndPostScriptClip
//----------------------------------------------------------------------------------------

void FW_CGraphicDevice::MacEndPostScriptClip()
{
	FW_ASSERT(fMacPostScriptClip);

	PicComment(kPostScriptBegin, 0,kODNULL);
	PicComment(kTextIsPostScript, 0,kODNULL);
	DrawString("\ppse currentpoint grestore moveto psb");
	PicComment(kPostScriptEnd, 0,kODNULL);

	fMacPostScriptClip = FALSE;
}
#endif

//----------------------------------------------------------------------------------------
//	FW_CGraphicDevice::SelectFont
//----------------------------------------------------------------------------------------

void FW_CGraphicDevice::SelectFont(const FW_PFont& font, FW_Boolean scale)
{
	FW_ASSERT(!fSuspended);
	FW_ASSERT(fOpenDeviceCount != 0);
	
	FW_CHECK_PLATFORM_CANVAS

	FW_ASSERT((const void*)font != NULL);

#ifdef FW_BUILD_WIN
	FW_CString255 fontName;
	font->GetFontName(fontName);
	fGDIFont.SetFontName(fontName);
	fGDIFont.SetFontStyle(font->GetFontStyle());
#endif
#ifdef FW_BUILD_MAC
	SetFontID(font->MacGetFontID());
	SetFontStyle(font->MacGetFontStyle());
#endif
	
	FW_CFixed fontSize = font->GetFontSize();
	if (scale)
	{
		FW_SPlatformPoint pt = fContext->LogicalToDevice(fontSize, fontSize);
		fontSize.SetInt(pt.Y());
	}

#ifdef FW_BUILD_WIN
	fGDIFont.SetFontSize(fontSize.AsInt());
	fGDIFont.SelectObject(fPlatformCanvas);
#endif		
#ifdef FW_BUILD_MAC
	SetFontSize(fontSize.AsInt());
	SetInGrafPort();
#endif
}

#ifdef FW_BUILD_WIN
//----------------------------------------------------------------------------------------
//	FW_CGraphicDevice::SelectInkAndFont
//----------------------------------------------------------------------------------------
//	Only call for Text shapes

void FW_CGraphicDevice::SelectInkAndFont(const FW_PInk& ink, const FW_PFont& font)
{
	SetTextColor(*ink->GetForeColor());
	SetBkColor(*ink->GetBackColor());
	SetBkMode(ink->GetTransferMode() == FW_kOr ? TRANSPARENT : OPAQUE);
	
	SelectFont(font, TRUE);
}
#endif

#ifdef FW_BUILD_WIN
//----------------------------------------------------------------------------------------
//	FW_CGraphicDevice::SetTextColor
//----------------------------------------------------------------------------------------

void FW_CGraphicDevice::SetTextColor(COLORREF textColor)
{
	FW_ASSERT(!fSuspended);
	FW_ASSERT(fOpenDeviceCount != 0);
	
	FW_CHECK_PLATFORM_CANVAS

	if (fTextColor != textColor)
	{
		fTextColor = textColor;
		::SetTextColor(fPlatformCanvas, textColor);
	}
}
#endif

#ifdef FW_BUILD_WIN
//----------------------------------------------------------------------------------------
//	FW_CGraphicDevice::SetBkColor
//----------------------------------------------------------------------------------------

void FW_CGraphicDevice::SetBkColor(COLORREF bkColor)
{
	FW_ASSERT(!fSuspended);
	FW_ASSERT(fOpenDeviceCount != 0);
	
	FW_CHECK_PLATFORM_CANVAS

	if (fBkColor != bkColor)
	{
		fBkColor = bkColor;
		::SetBkColor(fPlatformCanvas, bkColor);
	}
}
#endif

#ifdef FW_BUILD_WIN
//----------------------------------------------------------------------------------------
//	PrivWinROP2
//----------------------------------------------------------------------------------------

int static PrivWinROP2(FW_TransferModes transferMode)
{
	if ((transferMode & 0xFFFF0000L) == 0)
		return (int)transferMode;		// Native transfer mode
	else
	{
		FW_ASSERT(transferMode <= FW_LastTransferMode);
		return FW_gTransferModes[transferMode & 0x0000FFFFL];
	}
}
#endif

#ifdef FW_BUILD_WIN
//----------------------------------------------------------------------------------------
//	FW_CGraphicDevice::SelectInkAndStyle
//----------------------------------------------------------------------------------------
// Only call for shapeCategory == FW_kGeometricShape or FW_kGeometricShapeWithInvert

FW_Boolean FW_CGraphicDevice::SelectInkAndStyle(const FW_PInk& ink, 
												   const FW_PStyle& style,
												   FW_EShapeCategories shapeCategory, 
												   FW_ERenderVerbs renderVerb)
{
	FW_ASSERT(!fSuspended);
	FW_ASSERT(fOpenDeviceCount != 0);
	
	FW_CHECK_PLATFORM_CANVAS

	FW_ASSERT((const void*)ink != NULL);
	FW_ASSERT((const void*)style != NULL);
	
	FW_Boolean frameWithBrush = FALSE;
	
	// ----- Set transfer mode -----
	if (shapeCategory == FW_kGeometricShapeWithInvert && 
		ink->GetTransferMode() == FW_kInvert || 
		ink->GetTransferMode() == FW_kHilite)
	{
		SetDrawingMode(R2_COPYPEN);
	}
	else
		SetDrawingMode(::PrivWinROP2(ink->GetTransferMode()));

	// ----- Set pen size (even if filling)	
	FW_CFixed penSize = style->GetPenSize();
	if (!penSize)
		fPenSize.Set(1,1);
	else
		fPenSize = fContext->LogicalToDevice(penSize, penSize);

	// ----- Set pen and brush -----
	int nBkMode = OPAQUE;
	if (renderVerb == FW_kFrame)
	{
		FW_EStyleDash dash = style->GetDashStyle();
		frameWithBrush = (dash == FW_kSolidLine) && (!ink->PrivSpecialTransferMode()) && (!style->GetPattern()->IsBlack());
		
		if (frameWithBrush)
		{
			// NOTE: I don't set the Pen because we will not use it
			
			// ----- Set foreground and background Color -----
			SetTextColor(*ink->GetForeColor());
			SetBkColor(*ink->GetBackColor());
	
			// ----- Set Brush -----
			fGDIBrush.SetBrushPattern(style->GetPattern());
		}
		else
		{
			// ----- Set the Brush -----
			fGDIBrush.SetStockID(NULL_BRUSH);

			// ----- Set the pen -----
			fGDIPen.SetPenSize(fPenSize.y);
			fGDIPen.SetPenStyle(dash);
			fGDIPen.SetPenColor(ink->GetTransferMode() == FW_kErase ? *ink->GetBackColor() : *ink->GetForeColor());
			
			if ((dash & FW_kOpaque) == 0)
				nBkMode = TRANSPARENT;
			else
				SetBkColor(*ink->GetBackColor());
		}
	}
	else
	{
		// ----- Set the pen -----
		fGDIPen.SetStockID(NULL_PEN);

		// ----- Set the Brush -----
		FW_TransferModes transferMode = ink->GetTransferMode();
		if (transferMode == FW_kErase)
		{
			SetTextColor(FW_kRGBBlack);
			SetBkColor(FW_kRGBWhite);
			fGDIBrush.SetBrushColor(*ink->GetBackColor());
		}
		else if (transferMode == FW_kInvert || transferMode == FW_kHilite)
		{
			SetTextColor(FW_kRGBBlack);
			SetBkColor(FW_kRGBWhite);
			fGDIBrush.SetStockID(BLACK_BRUSH);
		}
		else
		{
			const FW_PPattern& pattern = style->GetPattern();

			if (pattern->IsBlack())
			{
				// If pattern is black let just use a solid brush
				SetTextColor(FW_kRGBBlack);
				SetBkColor(FW_kRGBWhite);
				fGDIBrush.SetBrushColor(*ink->GetForeColor());
			}
			else
			{
				SetTextColor(*ink->GetForeColor());
				SetBkColor(*ink->GetBackColor());
				fGDIBrush.SetBrushPattern(pattern);
			}
		}
	}
	
	SetBkMode(nBkMode);

	if (!frameWithBrush)
	{
		fGDIPen.SelectObject(fPlatformCanvas);
		fGDIBrush.SelectObject(fPlatformCanvas);
	}

	return frameWithBrush;
}
#endif
