//========================================================================================
//
//	File:				FWRecShp.cpp
//	Release Version:	$ 1.0d11 $
//
//	Copyright:	 1993, 1995 by Apple Computer, Inc., all rights reserved.
//
//========================================================================================

#include "FWOS.hpp"

#ifndef FWRECSHP_H
#include "FWRecShp.h"
#endif

#ifndef FWGC_H
#include "FWGC.h"
#endif

#ifndef FWGRGLOB_H
#include "FWGrGlob.h"
#endif

#ifndef FWGRUTIL_H
#include "FWGrUtil.h"
#endif

#ifndef FWFXMATH_H
#include "FWFxMath.h"
#endif

#ifndef FWRASTER_H
#include "FWRaster.h"
#endif

// ----- Foundation Includes -----

#ifndef FWSTREAM_H
#include "FWStream.h"
#endif

//========================================================================================
//	RunTime Info
//========================================================================================

#if FW_LIB_EXPORT_PRAGMAS
#pragma lib_export on
#endif

#ifdef FW_BUILD_MAC
#pragma segment FWGraphics_RectShape
#endif

FW_DEFINE_CLASS_M1(FW_CRectShape, FW_CBoundedShape)

FW_REGISTER_ARCHIVABLE_CLASS(FW_LRectShape, FW_CRectShape, FW_CRectShape::Read, FW_CShape::Write)

//========================================================================================
//	class FW_CRectShape
//========================================================================================

//----------------------------------------------------------------------------------------
//	FW_CRectShape::FW_CRectShape
//----------------------------------------------------------------------------------------

FW_CRectShape::FW_CRectShape() :
	FW_CBoundedShape(FW_kZeroRect, FW_kFill, FW_kNormalInk, FW_kNormalStyle, FW_kNormalFont)
{
	FW_END_CONSTRUCTOR
}

//----------------------------------------------------------------------------------------
//	FW_CRectShape::FW_CRectShape
//----------------------------------------------------------------------------------------

FW_CRectShape::FW_CRectShape(const FW_CRectShape& other) :
	FW_CBoundedShape(other)
{
	FW_END_CONSTRUCTOR
}

//----------------------------------------------------------------------------------------
//	FW_CRectShape::FW_CRectShape
//----------------------------------------------------------------------------------------

FW_CRectShape::FW_CRectShape(const FW_CRect& rect,
							 FW_ERenderVerbs renderVerb,
							 const FW_PInk& ink,
					 		 const FW_PStyle& style) :
	FW_CBoundedShape(rect, renderVerb, ink, style, FW_kNormalFont)
{
	FW_END_CONSTRUCTOR
}

//----------------------------------------------------------------------------------------
//	FW_CRectShape::FW_CRectShape
//----------------------------------------------------------------------------------------

FW_CRectShape::FW_CRectShape(FW_CReadableStream& archive) :
	FW_CBoundedShape(archive)
{
	FW_END_CONSTRUCTOR
}

//----------------------------------------------------------------------------------------
//	FW_CRectShape::~FW_CRectShape
//----------------------------------------------------------------------------------------

FW_CRectShape::~FW_CRectShape()
{
	FW_START_DESTRUCTOR
}

//----------------------------------------------------------------------------------------
//	FW_CRectShape::Render
//----------------------------------------------------------------------------------------

void FW_CRectShape::Render(FW_CGraphicContext& gc) const
{	
	gc.GetRasterizer()->RenderRect(gc,
									fRect,
									GetRenderVerb(),
									fInk,
									fStyle);
}

//----------------------------------------------------------------------------------------
//	FW_CRectShape::RenderRect
//----------------------------------------------------------------------------------------

void FW_CRectShape::RenderRect(FW_CGraphicContext& gc,
								const FW_CRect& rect, 
								FW_ERenderVerbs renderVerb, 
								const FW_PInk& ink,
								const FW_PStyle& style)
{	
	gc.GetRasterizer()->RenderRect(gc,
									rect,
									renderVerb,
									ink,
									style);
}

//----------------------------------------------------------------------------------------
//	FW_CRectShape::HitTest
//----------------------------------------------------------------------------------------

FW_Boolean FW_CRectShape::HitTest(FW_CGraphicContext& gc,
								  const FW_CPoint& test,
								  FW_CFixed tolerance) const
{	
	if (FW_CBoundedShape::HitTest(gc, test, tolerance))
	{
		if (fRenderVerb == FW_kFrame)
		{			
			FW_CRect bounds(fRect);
			FW_CFixed inset = tolerance + GetPenSize();
			bounds.Inset(inset, inset);
			return ! bounds.Contains(test);
		}
		
		return TRUE;
	}
	
	return FALSE;
}

//----------------------------------------------------------------------------------------
//	FW_CRectShape::Copy
//----------------------------------------------------------------------------------------

FW_CShape* FW_CRectShape::Copy() const
{
	return FW_NEW(FW_CRectShape, (*this));
}

//----------------------------------------------------------------------------------------
//	FW_CRectShape::Read
//----------------------------------------------------------------------------------------

void* FW_CRectShape::Read(FW_CReadableStream& archive)
{
	return FW_NEW(FW_CRectShape, (archive));
}
