//========================================================================================
//
//	File:				FWRegion.cpp
//	Release Version:	$ 1.0d11 $
//
//	Copyright:	 1993, 1995 by Apple Computer, Inc., all rights reserved.
//
//========================================================================================

#include "FWOS.hpp"

#ifndef FWREGION_H
#include "FWRegion.h"
#endif

#ifndef FWFXMATH_H
#include "FWFxMath.h"
#endif

#ifndef FWGRUTIL_H
#include "FWGrUtil.h"
#endif

#ifndef FWGRGLOB_H
#include "FWGrGlob.h"
#endif

#ifndef FWGRDEF_H
#include "FWGrDef.h"
#endif

// ----- Foundation Includes -----

#ifndef FWSTREAM_H
#include "FWStream.h"
#endif

#ifndef FWDEBUG_H
#include "FWDebug.h"
#endif

#ifndef FWMEMHLP_H
#include "FWMemHlp.h"
#endif

//========================================================================================
//	RunTime Info
//========================================================================================

#if FW_LIB_EXPORT_PRAGMAS
#pragma lib_export on
#endif

#ifdef FW_BUILD_MAC
#pragma segment FWGraphx_Region
#endif

//========================================================================================
//	Region utilities
//========================================================================================

//----------------------------------------------------------------------------------------
//	FW_NewRegion
//----------------------------------------------------------------------------------------

FW_FUNC_ATTR ODRgnHandle FW_NewRegion()
{	
#ifdef FW_BUILD_MAC
	return ::NewRgn();
#endif
#ifdef FW_BUILD_WIN
	return	::CreateRectRgn(0,0,0,0);
#endif
}

//----------------------------------------------------------------------------------------
//	FW_CreateRectRegion
//----------------------------------------------------------------------------------------

FW_FUNC_ATTR ODRgnHandle FW_CreateRectRegion(const FW_CRect& rect)
{
	FW_SPlatformRect plfmRect = rect;
	
#ifdef FW_BUILD_MAC
	ODRgnHandle rgn = ::NewRgn();
	::RectRgn(rgn, &plfmRect);
	return rgn;
#endif
#ifdef FW_BUILD_WIN
	return ::CreateRectRgnIndirect(&plfmRect);
#endif
}

//----------------------------------------------------------------------------------------
//	FW_CreateOvalRegion
//----------------------------------------------------------------------------------------

FW_FUNC_ATTR ODRgnHandle FW_CreateOvalRegion(const FW_CRect& rect)
{
	FW_SPlatformRect plfmRect(rect);

#ifdef FW_BUILD_MAC
	GrafPtr svPort = FW_PrivMacOpenRegion();
	::FrameOval(&plfmRect);
	return FW_PrivMacCloseRegion(svPort);
#endif

#ifdef FW_BUILD_WIN
	return ::CreateEllipticRgnIndirect(&plfmRect);
#endif
}

//----------------------------------------------------------------------------------------
//	FW_CreateRoundRectRegion
//----------------------------------------------------------------------------------------

FW_FUNC_ATTR ODRgnHandle FW_CreateRoundRectRegion(const FW_CRect& rect, const FW_CPoint& ovalSize)
{
	FW_SPlatformRect plfmRect(rect);
	FW_SPlatformPoint plfmPoint(ovalSize);

#ifdef FW_BUILD_MAC
	GrafPtr svPort = FW_PrivMacOpenRegion();
	::FrameRoundRect(&plfmRect, plfmPoint.X(), plfmPoint.Y());
	return ::FW_PrivMacCloseRegion(svPort);
#endif
#ifdef FW_BUILD_WIN
	return ::CreateRoundRectRgn(plfmRect.left, plfmRect.top, 
								plfmRect.right, plfmRect.bottom, 
								plfmPoint.X(), plfmPoint.Y());
#endif
}

//----------------------------------------------------------------------------------------
//	FW_CreateArcRegion
//----------------------------------------------------------------------------------------

FW_FUNC_ATTR ODRgnHandle FW_CreateArcRegion(const FW_CRect& rect, short startAngle, short arcAngle)
{
	FW_SPlatformRect plfmRect = rect;

	// Create an elliptic region
#ifdef FW_BUILD_MAC
	GrafPtr savePort = ::FW_PrivMacOpenRegion();
	::FrameOval(&plfmRect);
	ODRgnHandle hEllipticRgn = ::FW_PrivMacCloseRegion(savePort);
#endif
#ifdef FW_BUILD_WIN
	// Create an elliptic region
	HRGN hEllipticRgn = ::CreateEllipticRgnIndirect(&plfmRect);
#endif

	short angle0 = startAngle,
		  angle1 = startAngle + arcAngle * 1 / 4,
		  angle2 = startAngle + arcAngle * 2 / 4,
		  angle3 = startAngle + arcAngle * 3 / 4,
		  angle4 = startAngle + arcAngle;
		  
	short width = plfmRect.right - plfmRect.left;
	short height = plfmRect.bottom - plfmRect.top;
	FW_SPlatformRect polygonRect(plfmRect.left - width, plfmRect.top - height,
								 plfmRect.right + width, plfmRect.bottom + height);
	
	FW_SPlatformPoint pt[6];
	
	FW_PrivCalcArcPoints(polygonRect, angle0, pt[0]);
	FW_PrivCalcArcPoints(polygonRect, angle1, pt[1]);
	FW_PrivCalcArcPoints(polygonRect, angle2, pt[2]);
	FW_PrivCalcArcPoints(polygonRect, angle3, pt[3]);
	FW_PrivCalcArcPoints(polygonRect, angle4, pt[4]);
	
	pt[5].Set((plfmRect.left + plfmRect.right) / 2,
			  (plfmRect.top + plfmRect.bottom) / 2);

	// Create a polygon region
#ifdef FW_BUILD_MAC
	 savePort = ::FW_PrivMacOpenRegion();
	
	::MoveTo(pt[0].h, pt[0].v);
	for(short i = 1; i < 6; i ++)
		::LineTo(pt[i].h, pt[i].v);

	ODRgnHandle hPolygonRgn = ::FW_PrivMacCloseRegion(savePort);
#endif
#ifdef FW_BUILD_WIN
	ODRgnHandle hPolygonRgn = ::CreatePolygonRgn(pt, 6, WINDING);
#endif

	// Intersect the regions
	ODRgnHandle hArcRegion = ::FW_IntersectRegion(hEllipticRgn, hPolygonRgn);
	
	// Clean up
	::FW_DisposeRegion(hEllipticRgn);
	::FW_DisposeRegion(hPolygonRgn);
	
	// Return the result
	return hArcRegion;
}

//----------------------------------------------------------------------------------------
//	FW_CreatePolygonRegion
//----------------------------------------------------------------------------------------

FW_FUNC_ATTR ODRgnHandle FW_CreatePolygonRegion(unsigned long pointCount, const FW_CPoint* pointArray)
{
#ifdef FW_BUILD_MAC
	GrafPtr savePort = ::FW_PrivMacOpenRegion();
	
	FW_SPlatformPoint pt = pointArray[0];
	::MoveTo(pt.h, pt.v);
	
	for(unsigned long i = 1; i < pointCount; i ++)
	{
		pt = pointArray[i];
		::LineTo(pt.h, pt.v);
	}

	return ::FW_PrivMacCloseRegion(savePort);
#endif
#ifdef FW_BUILD_WIN
	FW_SPlatformPoint* points = new FW_SPlatformPoint[pointCount];
	
	for(unsigned long i = 0; i < pointCount; i ++)
		points[i] = pointArray[i];

	HRGN hRgn = ::CreatePolygonRgn(points, pointCount, ALTERNATE);
	delete[] points;
	
	return hRgn;
#endif
}

//----------------------------------------------------------------------------------------
//	FW_CreateLineRegion
//----------------------------------------------------------------------------------------

FW_FUNC_ATTR ODRgnHandle FW_CreateLineRegion(const FW_CPoint& startPt,
											 const FW_CPoint& endPt,
											 const FW_CPoint& lineThickness)
{
	ODRgnHandle rgn;
	
	FW_SPlatformPoint firstPoint(startPt);
	FW_SPlatformPoint lastPoint(endPt);
	FW_SPlatformPoint penSize(lineThickness);
	
	FW_SPlatformRect halfPen(-penSize.X() / 2, -penSize.Y() / 2, (penSize.X() + 1) / 2, (penSize.Y() + 1) / 2);

	FW_SPlatformRect rect(firstPoint.X() + halfPen.left, firstPoint.Y() + halfPen.top, firstPoint.X() + halfPen.right, firstPoint.Y() + halfPen.bottom);

	short regionCode = 0;
	if (lastPoint.X() < rect.left)
		regionCode = 0x0008;
	else if (lastPoint.X() > rect.right)
		regionCode = 0x0004;
	
	if (lastPoint.Y() < rect.top)
		regionCode |= 0x0001;
	else if (lastPoint.Y() > rect.bottom)
		regionCode |= 0x0002;

	short nbPoint;
	FW_SPlatformPoint pt[6];
	
	switch (regionCode)
	{
		case 0x0000:
			pt[0].Set(firstPoint.X() + halfPen.left, firstPoint.Y() + halfPen.top);
			pt[1].Set(lastPoint.X() + halfPen.right, lastPoint.Y() + halfPen.bottom);
			nbPoint = 2;
			break;
		case 0x0004:
			pt[0].Set(firstPoint.X() + halfPen.left, firstPoint.Y() + halfPen.top);
			pt[1].Set(lastPoint.X() + halfPen.right, lastPoint.Y() + halfPen.bottom);
			nbPoint = 2;
			break;
		case 0x0008:
			pt[0].Set(lastPoint.X() + halfPen.left, firstPoint.Y() + halfPen.top);
			pt[1].Set(firstPoint.X() + halfPen.right, lastPoint.Y() + halfPen.bottom);
			nbPoint = 2;
			break;
		case 0x0002:
			pt[0].Set(firstPoint.X() + halfPen.left, firstPoint.Y() + halfPen.top);
			pt[1].Set(lastPoint.X() + halfPen.right, lastPoint.Y() + halfPen.bottom);
			nbPoint = 2;
			break;
		case 0x0001:
			pt[0].Set(firstPoint.X() + halfPen.left, lastPoint.Y() + halfPen.top);
			pt[1].Set(lastPoint.X() + halfPen.right, firstPoint.Y() + halfPen.bottom);
			nbPoint = 2;
			break;
		case 0x0005:
			pt[0].Set(firstPoint.X() + halfPen.left, firstPoint.Y() + halfPen.top);
			pt[1].Set(firstPoint.X() + halfPen.left, firstPoint.Y() + halfPen.bottom);
			pt[2].Set(firstPoint.X() + halfPen.right, firstPoint.Y() + halfPen.bottom);
			pt[3].Set(lastPoint.X() + halfPen.right, lastPoint.Y() + halfPen.bottom);
			pt[4].Set(lastPoint.X() + halfPen.right, lastPoint.Y() + halfPen.top);
			pt[5].Set(lastPoint.X() + halfPen.left, lastPoint.Y() + halfPen.top);
			nbPoint = 6;
			break;
		case 0x0006:
			pt[0].Set(firstPoint.X() + halfPen.right, firstPoint.Y() + halfPen.top);
			pt[1].Set(firstPoint.X() + halfPen.left, firstPoint.Y() + halfPen.top);
			pt[2].Set(firstPoint.X() + halfPen.left, firstPoint.Y() + halfPen.bottom);
			pt[3].Set(lastPoint.X() + halfPen.left, lastPoint.Y() + halfPen.bottom);
			pt[4].Set(lastPoint.X() + halfPen.right, lastPoint.Y() + halfPen.bottom);
			pt[5].Set(lastPoint.X() + halfPen.right, lastPoint.Y() + halfPen.top);
			nbPoint = 6;
			break;
		case 0x000A:
			pt[0].Set(firstPoint.X() + halfPen.left, firstPoint.Y() + halfPen.top);
			pt[1].Set(firstPoint.X() + halfPen.right, firstPoint.Y() + halfPen.top);
			pt[2].Set(firstPoint.X() + halfPen.right, firstPoint.Y() + halfPen.bottom);
			pt[3].Set(lastPoint.X() + halfPen.right, lastPoint.Y() + halfPen.bottom);
			pt[4].Set(lastPoint.X() + halfPen.left, lastPoint.Y() + halfPen.bottom);
			pt[5].Set(lastPoint.X() + halfPen.left, lastPoint.Y() + halfPen.top);
			nbPoint = 6;
			break;
		case 0x0009:
			pt[0].Set(firstPoint.X() + halfPen.right, firstPoint.Y() + halfPen.top);
			pt[1].Set(firstPoint.X() + halfPen.right, firstPoint.Y() + halfPen.bottom);
			pt[2].Set(firstPoint.X() + halfPen.left, firstPoint.Y() + halfPen.bottom);
			pt[3].Set(lastPoint.X() + halfPen.left, lastPoint.Y() + halfPen.bottom);
			pt[4].Set(lastPoint.X() + halfPen.left, lastPoint.Y() + halfPen.top);
			pt[5].Set(lastPoint.X() + halfPen.right, lastPoint.Y() + halfPen.top);
			nbPoint = 6;
			break;
	}
	
#ifdef FW_BUILD_WIN
	if (nbPoint == 2)
		rgn = ::CreateRectRgn(pt[0].x, pt[0].y, pt[1].x, pt[1].y);
	else
		rgn = ::CreatePolygonRgn(pt, nbPoint, WINDING);
#endif		
	
#ifdef FW_BUILD_MAC
	if (nbPoint == 2)
	{
		rgn = ::NewRgn();
		FW_SPlatformRect rect(pt[0], pt[1]);
		::RectRgn(rgn, &rect);
	}
	else
	{
		GrafPtr svPort = ::FW_PrivMacOpenRegion();
		::MoveTo(pt[0].h, pt[0].v);
		::LineTo(pt[1].h, pt[1].v);
		::LineTo(pt[2].h, pt[2].v);			
		::LineTo(pt[3].h, pt[3].v);
		::LineTo(pt[4].h, pt[4].v);			
		::LineTo(pt[5].h, pt[5].v);
		::LineTo(pt[0].h, pt[0].v);
		rgn = ::FW_PrivMacCloseRegion(svPort);
	}
#endif

	return rgn;
}

//----------------------------------------------------------------------------------------
//	FW_CopyRegion
//----------------------------------------------------------------------------------------

FW_FUNC_ATTR ODRgnHandle FW_CopyRegion(ODRgnHandle scrRgn)
{
	FW_ASSERT(scrRgn != NULL);
	
#ifdef FW_BUILD_MAC
	ODRgnHandle newRgn = ::NewRgn();
	::CopyRgn(scrRgn, newRgn);
	return newRgn;
#endif
#ifdef FW_BUILD_WIN
	ODRgnHandle newRgn = ::CreateRectRgn(0,0,0,0);
	::CombineRgn(newRgn, scrRgn, NULL, RGN_COPY);
	return newRgn;
#endif
}

//----------------------------------------------------------------------------------------
//	FW_CopyRegion
//----------------------------------------------------------------------------------------

FW_FUNC_ATTR void FW_CopyRegion(ODRgnHandle scrRgn, ODRgnHandle dstRgn)
{
	FW_ASSERT(scrRgn != NULL);
	FW_ASSERT(dstRgn != NULL);
	
#ifdef FW_BUILD_MAC
	::CopyRgn(scrRgn, dstRgn);
#endif
#ifdef FW_BUILD_WIN
	::CombineRgn(dstRgn, scrRgn, NULL, RGN_COPY);
#endif
}

//----------------------------------------------------------------------------------------
//	FW_DisposeRegion
//----------------------------------------------------------------------------------------

FW_FUNC_ATTR void FW_DisposeRegion(ODRgnHandle rgn)
{
	FW_ASSERT(rgn);
	
#ifdef FW_BUILD_MAC
	::DisposeRgn(rgn);
#endif
#ifdef FW_BUILD_WIN
	::DeleteObject(rgn);
#endif
}

//----------------------------------------------------------------------------------------
//	FW_WriteRegion
//----------------------------------------------------------------------------------------

FW_FUNC_ATTR void FW_WriteRegion(ODRgnHandle rgnHandle, FW_CWritableStream& stream)
{
#ifdef FW_BUILD_WIN
	long size = ::GetRegionData(rgnHandle, 0, NULL);
	
	FW_CAcquireTemporarySystemHandle memHandle(size);
	LPRGNDATA rgnData = (LPRGNDATA) memHandle.GetPointer();
	::GetRegionData(rgnHandle, size, rgnData);
	
	stream << size;
	stream.Write(rgnData, size);
#endif
#ifdef FW_BUILD_MAC
	short size = (*rgnHandle)->rgnSize;
	
	FW_CAcquireLockedSystemHandle memHandle((FW_PlatformHandle)rgnHandle);
	void* rgnData = memHandle.GetPointer();
	
	stream << size;
	stream.Write(rgnData, size);
#endif
}

//----------------------------------------------------------------------------------------
//	FW_ReadRegion
//----------------------------------------------------------------------------------------

FW_FUNC_ATTR ODRgnHandle FW_ReadRegion(FW_CReadableStream& stream)
{
#ifdef FW_BUILD_WIN
	long size;
	stream >> size;
	
	FW_CAcquireTemporarySystemHandle memHandle(size);
	LPRGNDATA rgnData = (LPRGNDATA) memHandle.GetPointer();
	stream.Read(rgnData, size);

	return ::ExtCreateRegion(NULL, size, rgnData);
#endif
#ifdef FW_BUILD_MAC
	short size;
	stream >> size;
	
	FW_CAcquireTemporarySystemHandle memHandle(size);
	void* rgnData = memHandle.GetPointer();
	
	stream.Read(rgnData, size);

	memHandle.Orphan();
	return (ODRgnHandle) memHandle.GetPlatformHandle();
#endif
}

//----------------------------------------------------------------------------------------
//	FW_GetRegionBoundingBox
//----------------------------------------------------------------------------------------

FW_FUNC_ATTR void FW_GetRegionBoundingBox(ODRgnHandle rgn, FW_CRect& rect)
{
#ifdef FW_BUILD_MAC
	rect = (*rgn)->rgnBBox;
#endif
#ifdef FW_BUILD_WIN
	FW_SPlatformRect plfmRect;
	::GetRgnBox(rgn, &plfmRect);
	rect = plfmRect;
#endif
}

//----------------------------------------------------------------------------------------
//	FW_PointInRegion
//----------------------------------------------------------------------------------------

FW_FUNC_ATTR FW_Boolean FW_PointInRegion(ODRgnHandle rgn, const FW_CPoint& point)
{
	FW_SPlatformPoint plfmPt = point;
#ifdef FW_BUILD_MAC
	return ::PtInRgn(plfmPt, rgn);
#endif
#ifdef FW_BUILD_WIN
	return ::PtInRegion(rgn, plfmPt.X(), plfmPt.Y());
#endif
}

//----------------------------------------------------------------------------------------
//	FW_RectInRegion
//----------------------------------------------------------------------------------------

FW_FUNC_ATTR FW_Boolean FW_RectInRegion(ODRgnHandle rgn, const FW_CRect& rect)
{
	FW_SPlatformRect plfmRect = rect;
#ifdef FW_BUILD_MAC
	return ::RectInRgn(&plfmRect, rgn);
#endif
#ifdef FW_BUILD_WIN
	return ::RectInRegion(rgn, &plfmRect);
#endif
}

//----------------------------------------------------------------------------------------
//	FW_UnionRegion
//----------------------------------------------------------------------------------------

FW_FUNC_ATTR ODRgnHandle FW_UnionRegion(ODRgnHandle rgn1, ODRgnHandle rgn2)
{
	ODRgnHandle dstRgn = ::FW_NewRegion();
	
#ifdef FW_BUILD_MAC
	::UnionRgn(rgn1, rgn2, dstRgn);
#endif
#ifdef FW_BUILD_WIN
	::CombineRgn(dstRgn, rgn1, rgn2, RGN_OR);
#endif
	
	return dstRgn;
}

//----------------------------------------------------------------------------------------
//	FW_XorRegion
//----------------------------------------------------------------------------------------

FW_FUNC_ATTR ODRgnHandle FW_XorRegion(ODRgnHandle rgn1, ODRgnHandle rgn2)
{
	ODRgnHandle dstRgn = ::FW_NewRegion();
	
#ifdef FW_BUILD_MAC
	::XorRgn(rgn1, rgn2, dstRgn);
#endif
#ifdef FW_BUILD_WIN
	::CombineRgn(dstRgn, rgn1, rgn2, RGN_XOR);
#endif

	return dstRgn;
}

//----------------------------------------------------------------------------------------
//	FW_SubtractRegion
//----------------------------------------------------------------------------------------

FW_FUNC_ATTR ODRgnHandle FW_SubtractRegion(ODRgnHandle rgn1, ODRgnHandle rgn2)
{
	ODRgnHandle dstRgn = ::FW_NewRegion();
	
#ifdef FW_BUILD_MAC
	::DiffRgn(rgn1, rgn2, dstRgn);
#endif
#ifdef FW_BUILD_WIN
	::CombineRgn(dstRgn, rgn1, rgn2, RGN_DIFF);
#endif

	return dstRgn;
}

//----------------------------------------------------------------------------------------
//	FW_IntersectRegion
//----------------------------------------------------------------------------------------

FW_FUNC_ATTR ODRgnHandle FW_IntersectRegion(ODRgnHandle rgn1, ODRgnHandle rgn2)
{
	ODRgnHandle dstRgn = ::FW_NewRegion();
	
#ifdef FW_BUILD_MAC
	::SectRgn(rgn1, rgn2, dstRgn);
#endif
#ifdef FW_BUILD_WIN
	::CombineRgn(dstRgn, rgn1, rgn2, RGN_AND);
#endif

	return dstRgn;
}

//----------------------------------------------------------------------------------------
//	FW_OutlineRegion
//----------------------------------------------------------------------------------------

FW_FUNC_ATTR void FW_OutlineRegion(ODRgnHandle rgn, FW_CFixed outlineSize)
{
	ODRgnHandle tempRgn = ::FW_CopyRegion(rgn);
	
	if (outlineSize <= FW_kFixed0)
		outlineSize = FW_kFixedPos1;
	
	::FW_InsetRegion(tempRgn, outlineSize, outlineSize);
	
	::FW_SubtractRegion(rgn, tempRgn, rgn);

	::FW_DisposeRegion(tempRgn);
}

#ifdef FW_BUILD_WIN

//----------------------------------------------------------------------------------------
//	FW_PrivWinInsetRegionPoint
//----------------------------------------------------------------------------------------

static void FW_PrivWinInsetRegionRectangle(LPPOINT lppt, int dx, int dy, const POINT& ptCenter)
{
	if(lppt->x < ptCenter.x)
	{
		lppt->x += dx;
		if(lppt->x > ptCenter.x)
			lppt->x = ptCenter.x;
	}
	else
	{
		lppt->x -= dx;
		if(lppt->x < ptCenter.x)
			lppt->x = ptCenter.x;
	}

	if(lppt->y < ptCenter.y)
	{
		lppt->y += dy;
		if(lppt->y > ptCenter.y)
			lppt->y = ptCenter.y;
	}
	else
	{
		lppt->y -= dy;
		if(lppt->y < ptCenter.y)
			lppt->y = ptCenter.y;
	}
}

//----------------------------------------------------------------------------------------
//	FW_PrivWinInsetRegionRectangle
//----------------------------------------------------------------------------------------

static void FW_PrivWinInsetRegionRectangle(LPRECT lpRect, int dx, int dy, const POINT& ptCenter)
{
	FW_PrivWinInsetRegionRectangle(((LPPOINT) lpRect) + 0, dx, dy, ptCenter);
	FW_PrivWinInsetRegionRectangle(((LPPOINT) lpRect) + 1, dx, dy, ptCenter);
}

#endif

//----------------------------------------------------------------------------------------
//	FW_InsetRegion
//----------------------------------------------------------------------------------------

FW_FUNC_ATTR void FW_InsetRegion(ODRgnHandle rgn, FW_CFixed x, FW_CFixed y)
{
#ifdef FW_BUILD_MAC
	::InsetRgn(rgn, x.AsInt(), y.AsInt());
#endif
#ifdef FW_BUILD_WIN
	// Note: This code relies on partially documented Windows GDI data structures!

	// Get the current region data
	long size = ::GetRegionData(rgn, 0, NULL);
	
	FW_CAcquireTemporarySystemHandle memHandle(size);
	LPRGNDATA rgnData = (LPRGNDATA) memHandle.GetPointer();
	::GetRegionData(rgn, size, rgnData);
	
	// Mess with region data
	int dx = x.AsInt();
	int dy = y.AsInt();
	
	if(rgnData->rdh.iType == RDH_RECTANGLES)
	{
		POINT ptCenter;
		ptCenter.x = (rgnData->rdh.rcBound.left + rgnData->rdh.rcBound.right) / 2;
		ptCenter.y = (rgnData->rdh.rcBound.top  + rgnData->rdh.rcBound.bottom) / 2;
		
		LPRECT lpRect = (LPRECT) rgnData->Buffer;

		for(unsigned long i = 0; i < rgnData->rdh.nCount; ++ i, ++ lpRect)
			::FW_PrivWinInsetRegionRectangle(lpRect, dx, dy, ptCenter);
	}

	// Create a new region and replace the one passed in	
	HRGN hNewRgn = ::ExtCreateRegion(NULL, size, rgnData);
	if(hNewRgn == NULL)
		FW_Failure(FW_xMemoryExhausted);

	::FW_CopyRegion(hNewRgn, rgn);
	::FW_DisposeRegion(hNewRgn);
#endif
}

//----------------------------------------------------------------------------------------
//	FW_MapRegion
//----------------------------------------------------------------------------------------

FW_FUNC_ATTR void FW_MapRegion(ODRgnHandle rgn, const FW_CRect& srcRect, const FW_CRect& dstRect)
{
	FW_SPlatformRect plfmSrc(srcRect);
	FW_SPlatformRect plfmDst(dstRect);
	
#ifdef FW_BUILD_MAC
	::MapRgn(rgn, &plfmSrc, &plfmDst);
#endif
#ifdef FW_BUILD_WIN
	// Get the current region data
	long size = ::GetRegionData(rgn, 0, NULL);
	
	FW_CAcquireTemporarySystemHandle memHandle(size);
	LPRGNDATA rgnData = (LPRGNDATA) memHandle.GetPointer();
	::GetRegionData(rgn, size, rgnData);

	// Figure out the desired transform
	XFORM xform;
	xform.eM11 = (dstRect.Width() / srcRect.Width()).AsDouble();
	xform.eM12 = 0;
	xform.eM21 = 0;
	xform.eM22 = (dstRect.Height() / srcRect.Height()).AsDouble();
	xform.eDx  = (dstRect.left - srcRect.left).AsDouble();
	xform.eDy  = (dstRect.top  - srcRect.top).AsDouble();

	// Create a new region and replace the one passed in	
	HRGN hNewRgn = ::ExtCreateRegion(&xform, size, rgnData);
	if(hNewRgn == NULL)
		FW_Failure(FW_xMemoryExhausted);

	::FW_CopyRegion(hNewRgn, rgn);
	::FW_DisposeRegion(hNewRgn);
#endif
}

#ifdef FW_BUILD_MAC
//----------------------------------------------------------------------------------------
//	FW_PrivMacOpenRegion
//----------------------------------------------------------------------------------------

GrafPtr FW_PrivMacOpenRegion()
{
	GrafPtr svPort;
	::GetPort(&svPort);
	::SetPort(FW_gScratchWindow);
	::OpenRgn();
	return svPort;
}
#endif

#ifdef FW_BUILD_MAC
//----------------------------------------------------------------------------------------
//	FW_PrivMacCloseRegion
//----------------------------------------------------------------------------------------

ODRgnHandle FW_PrivMacCloseRegion(GrafPtr svPort)
{
	ODRgnHandle hRgn = ::NewRgn();
	::CloseRgn(hRgn);
	::SetPort(svPort);
	return hRgn;
}
#endif
