//========================================================================================
//
//	File:				FWRgnShp.cpp
//	Release Version:	$ 1.0d11 $
//
//	Copyright:	 1993, 1995 by Apple Computer, Inc., all rights reserved.
//
//========================================================================================

#include "FWOS.hpp"

#ifndef FWRGNSHP_H
#include "FWRgnShp.h"
#endif

#ifndef FWGRGLOB_H
#include "FWGrGlob.h"
#endif

#ifndef FWGC_H
#include "FWGC.h"
#endif

#ifndef FWGRUTIL_H
#include "FWGrUtil.h"
#endif

#ifndef FWFXMATH_H
#include "FWFxMath.h"
#endif

#ifndef FWRASTER_H
#include "FWRaster.h"
#endif

#ifndef FWREGION_H
#include "FWRegion.h"
#endif

#ifndef FWODGEOM_H
#include "FWODGeom.h"
#endif

// ----- Foundation Includes -----

#ifndef FWSTREAM_H
#include "FWStream.h"
#endif

// ----- OpenDoc Includes -----

#ifndef SOM_ODTransform_xh
#include <Trnsform.xh>
#endif

#ifndef FW_BUILD_WIN
#ifndef SOM_Module_OpenDoc_Polygon_defined
#include <Polygon.xh>
#endif
#endif

//========================================================================================
//	RunTime Info
//========================================================================================

#if FW_LIB_EXPORT_PRAGMAS
#pragma lib_export on
#endif

#ifdef FW_BUILD_MAC
#pragma segment FWGraphics_RegionShape
#endif

FW_DEFINE_CLASS_M1(FW_CRegionShape, FW_CShape)

FW_REGISTER_ARCHIVABLE_CLASS(FW_LRegionShape, FW_CRegionShape, FW_CRegionShape::Read, FW_CShape::Write)

//========================================================================================
//	class FW_CRegionShape
//========================================================================================

//----------------------------------------------------------------------------------------
//	FW_CRegionShape::FW_CRegionShape
//----------------------------------------------------------------------------------------

FW_CRegionShape::FW_CRegionShape(ODShape* odShape,
								 FW_ERenderVerbs renderVerb,
								 const FW_PInk& ink,
								 const FW_PStyle& style) :
	FW_CShape(renderVerb, ink, style, FW_kNormalFont),
	fODShape(odShape)
{
	FW_END_CONSTRUCTOR
}

//----------------------------------------------------------------------------------------
//	FW_CRegionShape::FW_CRegionShape
//----------------------------------------------------------------------------------------

FW_CRegionShape::FW_CRegionShape(const FW_CRegionShape& other) :
	FW_CShape(other),
	fODShape(NULL)
{
	if (other.fODShape)
		fODShape = other.fODShape->Copy(somGetGlobalEnvironment());		// [HLX] for now
		
	FW_END_CONSTRUCTOR
}

//----------------------------------------------------------------------------------------
//	FW_CRegionShape::FW_CRegionShape
//----------------------------------------------------------------------------------------

FW_CRegionShape::FW_CRegionShape(FW_CReadableStream& archive) :
	FW_CShape(archive),
	fODShape(NULL)
{
#ifndef FW_BUILD_WIN
	unsigned long dataSize;
	archive >> dataSize;
	ODPolygon* polygon = (ODPolygon*)::operator new(dataSize);
	archive.Read((char*)polygon, dataSize);
	
	Environment* ev = somGetGlobalEnvironment();	// [HLX] for now
	// [B1 Conversion] Should diseapear when region shape doesn't use ODShape
	fODShape = new ODShape;
	fODShape->SetPolygon(ev, polygon);
#else
	FW_DEBUG_MESSAGE("Not Yet Implemented");
#endif

	FW_END_CONSTRUCTOR
}

//----------------------------------------------------------------------------------------
//	FW_CRegionShape::~FW_CRegionShape
//----------------------------------------------------------------------------------------

FW_CRegionShape::~FW_CRegionShape()
{
	FW_START_DESTRUCTOR
	Environment* ev = somGetGlobalEnvironment();	// [HLX] for now
	DisposeODShape(ev);
}

//----------------------------------------------------------------------------------------
//	FW_CRegionShape::operator=
//----------------------------------------------------------------------------------------

FW_CRegionShape& FW_CRegionShape::operator=(const FW_CRegionShape& other)
{
	if (this != &other)
	{
		FW_CShape::operator=(other);
		fODShape = other.fODShape->Copy(somGetGlobalEnvironment());
	}
	
	return *this;
}

//----------------------------------------------------------------------------------------
//	FW_CRegionShape::Render
//----------------------------------------------------------------------------------------

void FW_CRegionShape::Render(FW_CGraphicContext& gc) const
{
	if (fODShape == NULL)
		return;

	gc.GetRasterizer()->RenderRegion(gc,
									fODShape,
									GetRenderVerb(),
									fInk,
									fStyle);
}

//----------------------------------------------------------------------------------------
//	FW_CRegionShape::Render
//----------------------------------------------------------------------------------------

void FW_CRegionShape::RenderRegion(FW_CGraphicContext& gc,
									ODShape* odShape, 
									FW_ERenderVerbs renderVerb, 
									const FW_PInk& ink,
									const FW_PStyle& style)
{
	FW_ASSERT(odShape != NULL);
	
	gc.GetRasterizer()->RenderRegion(gc,
									odShape,
									renderVerb,
									ink,
									style);
}

//----------------------------------------------------------------------------------------
//	FW_CRegionShape::AdoptODShape
//----------------------------------------------------------------------------------------

void FW_CRegionShape::AdoptODShape(ODShape* odShape)
{
	FW_ASSERT(odShape != NULL);
	
	Environment* ev = somGetGlobalEnvironment();	// [HLX] for now
	DisposeODShape(ev);
	fODShape = odShape;	
}

//----------------------------------------------------------------------------------------
//	FW_CRegionShape::OrphanODShape
//----------------------------------------------------------------------------------------

ODShape* FW_CRegionShape::OrphanODShape()
{
	ODShape* odTempShape = fODShape;
	fODShape = NULL;
	return odTempShape;
}

//----------------------------------------------------------------------------------------
//	FW_CRegionShape::SetGeometry
//----------------------------------------------------------------------------------------

void FW_CRegionShape::SetGeometry(ODShape* odShape)
{
	Environment* ev = somGetGlobalEnvironment();	// [HLX] for now
	DisposeODShape(ev);
	fODShape = odShape->Copy(ev);
}

//----------------------------------------------------------------------------------------
//	FW_CRegionShape::DisposeODShape
//----------------------------------------------------------------------------------------

void FW_CRegionShape::DisposeODShape(Environment* ev)
{
	fODShape->Release(ev);
	fODShape = NULL;
}

//----------------------------------------------------------------------------------------
//	FW_CRegionShape::HitTest
//----------------------------------------------------------------------------------------

FW_Boolean FW_CRegionShape::HitTest(FW_CGraphicContext& gc,
									const FW_CPoint& test,
									FW_CFixed tolerance) const
{
	if (fODShape == NULL)
		return FALSE;

	if(fRenderVerb == FW_kNoRendering)
		return FALSE;

	Environment* ev = gc.GetEnvironment();

	ODRgnHandle rgn = ::FW_GetShapeRegion(ev, fODShape);

	FW_CRect testRect(test.x - tolerance, test.y - tolerance,
					  test.x + tolerance, test.y + tolerance);
	
	if (!::FW_RectInRegion(rgn, testRect))
		return FALSE;

	if (fRenderVerb == FW_kFrame)
	{
		ODShape* insideShape = fODShape->Copy(ev);
		::FW_OutlineODShape(ev, insideShape, GetPenSize());

		ODRgnHandle insideRgn = ::FW_GetShapeRegion(ev, insideShape);
		
		FW_Boolean isInside = ::FW_RectInRegion(insideRgn, testRect);

		insideShape->Release(ev);
		
		return !isInside;
	}
	
	return TRUE;
}

//----------------------------------------------------------------------------------------
//	FW_CRegionShape::Copy
//----------------------------------------------------------------------------------------

FW_CShape* FW_CRegionShape::Copy() const
{
	return FW_NEW(FW_CRegionShape, (*this));
}

//----------------------------------------------------------------------------------------
//	FW_CRegionShape::Flatten
//----------------------------------------------------------------------------------------

void FW_CRegionShape::Flatten(FW_CWritableStream& archive) const
{
	if (fODShape == NULL)
		return;

	FW_DEBUG_MESSAGE("Not Yet Implemented");

/*
#ifndef FW_BUILD_WIN
	ODPolygon* polygon = fODShape->CopyPolygon(somGetGlobalEnvironment());
	
	unsigned long dataSize = sizeof(ODSLong) + polygon->nContours * sizeof(ODSLong);
	x
	const ODContour *c = &polygon->firstContour;
	for( ODSLong i = polygon->nContours; i>0; i-- ) {
		dataSize += c->nVertices * sizeof(ODPoint);
		c = (ODContour*)&c->vertex[c->nVertices];	// give the next contour
	}

	archive << dataSize;
	archive.Write((char*)polygon, dataSize);
	
	delete polygon;	// [HLX] can I delete that
#else	
FW_UNUSED(archive);
	FW_DEBUG_MESSAGE("Not Yet Implemented");
#endif
*/
}

//----------------------------------------------------------------------------------------
//	FW_CRegionShape::Read
//----------------------------------------------------------------------------------------

void* FW_CRegionShape::Read(FW_CReadableStream& archive)
{
	return FW_NEW(FW_CRegionShape, (archive));
}

//----------------------------------------------------------------------------------------
//	FW_CRegionShape::GetBounds
//----------------------------------------------------------------------------------------

void FW_CRegionShape::GetBounds(FW_CGraphicContext& gc, FW_CRect& rect) const
{
FW_UNUSED(gc);

	fODShape->GetBoundingBox(gc.GetEnvironment(), (ODRect*) &rect);
}

//----------------------------------------------------------------------------------------
//	FW_CRegionShape::Transform
//----------------------------------------------------------------------------------------

void FW_CRegionShape::Transform(Environment* ev, ODTransform* odTransform)
{
	fODShape->Transform(ev, odTransform);
}

//----------------------------------------------------------------------------------------
//	FW_CRegionShape::InverseTransform
//----------------------------------------------------------------------------------------

void FW_CRegionShape::InverseTransform(Environment* ev, ODTransform* odTransform)
{
	fODShape->InverseTransform(ev, odTransform);
}

//----------------------------------------------------------------------------------------
//	FW_CRegionShape::MoveShape
//----------------------------------------------------------------------------------------

void FW_CRegionShape::MoveShape(FW_CFixed deltaX, FW_CFixed deltaY)
{
	Environment* ev = somGetGlobalEnvironment();	// [HLX] for now
		
	ODTransform* odTransform = ::FW_NewODTransform(ev, FW_CPoint(deltaX, deltaY));
	fODShape->Transform(ev, odTransform);
	odTransform->Release(ev);
}

//----------------------------------------------------------------------------------------
//	FW_CRegionShape::MoveShapeTo
//----------------------------------------------------------------------------------------

void FW_CRegionShape::MoveShapeTo(FW_CFixed x, FW_CFixed y)
{
	Environment* ev = somGetGlobalEnvironment();	// [HLX] for now

	FW_CRect box;
	fODShape->GetBoundingBox(ev, (ODRect*) &box);
	MoveShape(x - box.left, y - box.top);
}

//----------------------------------------------------------------------------------------
//	FW_CRegionShape::Inset
//----------------------------------------------------------------------------------------

void FW_CRegionShape::Inset(FW_CFixed x, FW_CFixed y)
{
	Environment* ev = somGetGlobalEnvironment();	// [HLX] for now

	ODRgnHandle plfmRgn = ::FW_CopyRegion(::FW_GetShapeRegion(ev, fODShape));
	
	::FW_InsetRegion(plfmRgn, x, y);

	::FW_SetShapeRegion(ev, fODShape, plfmRgn);
	::FW_DisposeRegion(plfmRgn);
}

