//========================================================================================
//
//	File:				FWStyle.cpp
//	Release Version:	$ 1.0d11 $
//
//	Copyright:	 1993, 1995 by Apple Computer, Inc., all rights reserved.
//
//========================================================================================

#include "FWOS.hpp"

#if defined(FW_PROFILER_CALLS) && defined(__MWERKS__)
#pragma profile on
#endif

#ifndef FWSTYLE_H
#include "FWStyle.h"
#endif

#ifndef FWGC_H
#include "FWGC.h"
#endif

#ifndef FWPOINT_H
#include "FWPoint.h"
#endif

#ifndef FWFXMATH_H
#include "FWFxMath.h"
#endif

#ifndef FWGRUTIL_H
#include "FWGrUtil.h"
#endif

#ifndef FWBWPAT_H
#include "FWBWPat.h"
#endif

// ----- Foundation Includes -----

#ifndef FWSTREAM_H
#include "FWStream.h"
#endif

// ----- OpenDoc Includes -----

#ifndef _TRANSFORM_
#include "Trnsform.xh"
#endif

// ----- Macintosh Includes -----

#if defined(FW_BUILD_MAC) && !defined(__TEXTUTILS__)
#include <TextUtils.h>
#endif

#if defined(FW_BUILD_MAC) && !defined(__FONTS__)
#include <Fonts.h>
#endif

//========================================================================================
//	RunTime Info
//========================================================================================

#if FW_LIB_EXPORT_PRAGMAS
#pragma lib_export on
#endif

#ifdef FW_BUILD_MAC
#pragma segment FWGraphx_Style
#endif

FW_DEFINE_CLASS_M1(FW_PStyle, FW_CGraphicCountedPtr)
FW_DEFINE_CLASS_M1(FW_CStyleRep, FW_CGraphicCountedPtrRep)

FW_REGISTER_ARCHIVABLE_CLASS(FW_LStyleRep, FW_CStyleRep, FW_CStyleRep::Read, FW_CGraphicCountedPtrRep::Write)

//========================================================================================
//	constants
//========================================================================================

FW_CFixed FW_kDefaultPenSize = FW_IntToFixed(0);

static FW_CStyleRep* PrivCreateStandardStyle(FW_EStandardStyles std)
{
	FW_EStyleDash dash;

	switch(std)
	{
	default:
		FW_ASSERT(FALSE);		// Invalid standard style
								// fall-through
	case FW_kNormalStyle:
		dash = FW_kSolidLine;
		break;
 
	case FW_kDashStyle:
		dash = FW_kDash;
		break;
 
	case FW_kDotStyle:
		dash = FW_kDot;
		break;
 
	case FW_kDashDotStyle:
		dash = FW_kDashDot;
		break;
 
	case FW_kDashDotDotStyle:
		dash = FW_kDashDotDot;
		break;
	}

	return new FW_CStyleRep(FW_kDefaultPenSize, dash, FW_kBlackPat);
}

//========================================================================================
//	class FW_PStyle
//========================================================================================

//----------------------------------------------------------------------------------------
//	FW_PStyle::FW_PStyle
//----------------------------------------------------------------------------------------

FW_PStyle::FW_PStyle() :
	FW_CGraphicCountedPtr()
{
}

//----------------------------------------------------------------------------------------
//	FW_PStyle::FW_PStyle
//----------------------------------------------------------------------------------------

FW_PStyle::FW_PStyle(FW_CFixed penSize, FW_EStyleDash dash) :
	FW_CGraphicCountedPtr()
{
	SetRep(new FW_CStyleRep(penSize, dash, FW_kBlackPat));
}

//----------------------------------------------------------------------------------------
//	FW_PStyle::FW_PStyle
//----------------------------------------------------------------------------------------

FW_PStyle::FW_PStyle(FW_CFixed penSize, const FW_PPattern& pattern) :
	FW_CGraphicCountedPtr()
{
	SetRep(new FW_CStyleRep(penSize, FW_kSolidLine, pattern));
}

//----------------------------------------------------------------------------------------
//	FW_PStyle::FW_PStyle
//----------------------------------------------------------------------------------------

FW_PStyle::FW_PStyle(const FW_PStyle& other) :
	FW_CGraphicCountedPtr(other)
{
}

//----------------------------------------------------------------------------------------
//	FW_PStyle::FW_PStyle
//----------------------------------------------------------------------------------------

FW_PStyle::~FW_PStyle()
{
}

//----------------------------------------------------------------------------------------
//	FW_PStyle::operator=
//----------------------------------------------------------------------------------------

FW_PStyle& FW_PStyle::operator=(const FW_PStyle& other)
{
	// We don't need to test this == &other because SetRep will do it
	SetRep(other.GetRep());
	return *this;
}

//----------------------------------------------------------------------------------------
//	FW_PStyle::operator=
//----------------------------------------------------------------------------------------

FW_PStyle::FW_PStyle(FW_EStandardStyles std)
{
	SetRep(PrivCreateStandardStyle(std));
}

//----------------------------------------------------------------------------------------
//	FW_PStyle::operator=
//----------------------------------------------------------------------------------------

FW_PStyle& FW_PStyle::operator=(FW_EStandardStyles std)
{
	SetRep(PrivCreateStandardStyle(std));
	return *this;
}

//========================================================================================
//	class FW_CStyleRep
//========================================================================================

//----------------------------------------------------------------------------------------
//	FW_CStyleRep::FW_CStyleRep
//----------------------------------------------------------------------------------------

FW_CStyleRep::FW_CStyleRep(FW_CFixed penSize, FW_EStyleDash dash, const FW_PPattern& pattern) :
	fPenSize(penSize),
	fDash(dash),
	fPattern(pattern)
{
}

//----------------------------------------------------------------------------------------
//	FW_CStyleRep::~FW_CStyleRep
//----------------------------------------------------------------------------------------

FW_CStyleRep::~FW_CStyleRep()
{	
}

//----------------------------------------------------------------------------------------
//	FW_CStyleRep::FW_CStyleRep
//----------------------------------------------------------------------------------------

FW_CStyleRep& FW_CStyleRep::operator=(const FW_CStyleRep& otherRep)
{
	fPenSize = otherRep.fPenSize;
	fDash    = otherRep.fDash;
	fPattern = otherRep.fPattern;
	return *this;
}

//----------------------------------------------------------------------------------------
//	FW_CStyleRep::Copy
//----------------------------------------------------------------------------------------

FW_PPattern& FW_CStyleRep::GetUnSharedPattern()
{
	if (fPattern->GetRefCount() > 1)
		fPattern = fPattern->Copy();
		
	return fPattern;
}

//----------------------------------------------------------------------------------------
//	FW_CStyleRep::Copy
//----------------------------------------------------------------------------------------

FW_PStyle FW_CStyleRep::Copy() const
{
	if (fDash == FW_kSolidLine)
		return FW_PStyle(fPenSize, fPattern);
	
	return FW_PStyle(fPenSize, fDash);
}

//----------------------------------------------------------------------------------------
//	FW_CStyleRep::Flatten
//----------------------------------------------------------------------------------------

void FW_CStyleRep::Flatten(FW_CWritableStream& archive) const
{
	archive << fPenSize;
	archive << (short) fDash;
	archive << fPattern;
}

//----------------------------------------------------------------------------------------
//	FW_CStyleRep::IsEqual
//----------------------------------------------------------------------------------------

FW_Boolean FW_CStyleRep::IsEqual(const FW_CGraphicCountedPtrRep* other) const
{
	if (other == this)
		return TRUE;
		
	FW_CStyleRep* styleRep = FW_DYNAMIC_CAST(FW_CStyleRep, other);
	if (styleRep != NULL)
	{
		return fPenSize == styleRep->fPenSize &&
				fDash == styleRep->fDash &&
				fPattern == styleRep->fPattern;
	}
	
	return FALSE;
}

//----------------------------------------------------------------------------------------
//	FW_CStyleRep::Read
//----------------------------------------------------------------------------------------

void* FW_CStyleRep::Read(FW_CReadableStream& archive)
{
	FW_CFixed penSize;
	short dash;
	FW_PPattern pattern;
	
	archive >> penSize;
	archive >> dash;
	archive >> pattern;

	return new FW_CStyleRep(penSize, (FW_EStyleDash) dash, pattern);
}
