//========================================================================================
//
//	File:				FWTxtShp.cpp
//	Release Version:	$ 1.0d11 $
//
//	Copyright:	 1993, 1995 by Apple Computer, Inc., all rights reserved.
//
//========================================================================================

#include "FWOS.hpp"

#ifndef FWTXTSHP_H
#include "FWTxtShp.h"
#endif

#ifndef FWTXTBUF_H
#include "FWTxtBuf.h"
#endif

#ifndef FWRASTER_H
#include "FWRaster.h"
#endif

// ----- Foundation Includes -----

#ifndef FWSTREAM_H
#include "FWStream.h"
#endif

// ----- OpenDoc Includes -----

#ifndef _TRANSFORM_
#include <Trnsform.xh>
#endif

#if FW_LIB_EXPORT_PRAGMAS
#pragma lib_export on
#endif

//========================================================================================
//	RunTime Info
//========================================================================================

#ifdef FW_BUILD_MAC
#pragma segment FWGraphics_BaseTextShape
#endif

FW_DEFINE_CLASS_M1(FW_CTextShape, FW_CBaseTextShape)

FW_REGISTER_ARCHIVABLE_CLASS(FW_LTextShape, FW_CTextShape, FW_CTextShape::Read, FW_CShape::Write)

//========================================================================================
//	class FW_CTextShape
//========================================================================================

//----------------------------------------------------------------------------------------
//	FW_CTextShape::FW_CTextShape
//----------------------------------------------------------------------------------------

FW_CTextShape::FW_CTextShape(const FW_CTextShape& other) :
	FW_CBaseTextShape(other),
	fPosition(other.fPosition),
	fTextAlignment(other.fTextAlignment)
{
	FW_END_CONSTRUCTOR
}

//----------------------------------------------------------------------------------------
//	FW_CTextShape::FW_CTextShape
//----------------------------------------------------------------------------------------

FW_CTextShape::FW_CTextShape(const FW_CString& string, 
							FW_CFixed xPos,
							FW_CFixed yPos,
							const FW_PFont& font,
							FW_TextAlignment textAlignment,
							const FW_PInk& ink) :
	FW_CBaseTextShape(string, ink, font),
	fPosition(xPos, yPos),
	fTextAlignment(textAlignment)
{
	FW_END_CONSTRUCTOR
}

//----------------------------------------------------------------------------------------
//	FW_CTextShape::FW_CTextShape
//----------------------------------------------------------------------------------------

FW_CTextShape::FW_CTextShape() :
	FW_CBaseTextShape(),
	fTextAlignment(kDefaultAlignment)
		// fPosition is set to 0's by the default constructor
{
	FW_END_CONSTRUCTOR
}

//----------------------------------------------------------------------------------------
//	FW_CTextShape::FW_CTextShape
//----------------------------------------------------------------------------------------

FW_CTextShape::FW_CTextShape(FW_CTextReader* textReader, 
							FW_CFixed xPos, FW_CFixed yPos,
							const FW_PFont& font,
							FW_TextAlignment textAlignment,
							const FW_PInk& ink) :
	FW_CBaseTextShape(textReader, ink, font),
	fPosition(xPos, yPos),
	fTextAlignment(textAlignment)
{
	FW_END_CONSTRUCTOR
}

//----------------------------------------------------------------------------------------
//	FW_CTextShape::FW_CTextShape
//----------------------------------------------------------------------------------------

FW_CTextShape::FW_CTextShape(FW_CReadableStream& archive) :
	FW_CBaseTextShape(archive)
{
	archive >> fPosition;
	archive >> fTextAlignment;

	FW_END_CONSTRUCTOR
}

//----------------------------------------------------------------------------------------
//	FW_CTextShape::~FW_CTextShape
//----------------------------------------------------------------------------------------

FW_CTextShape::~FW_CTextShape()
{
	FW_START_DESTRUCTOR
}

//----------------------------------------------------------------------------------------
//	FW_CTextShape::operator=
//----------------------------------------------------------------------------------------

FW_CTextShape& FW_CTextShape::operator=(const FW_CTextShape& other)
{
	if (this != &other)
	{
		FW_CBaseTextShape::operator=(other);
		
		fPosition = other.fPosition;
		fTextAlignment = other.fTextAlignment;
	}
	
	return *this;
}

//----------------------------------------------------------------------------------------
//	FW_CTextShape::Render
//----------------------------------------------------------------------------------------

void FW_CTextShape::Render(FW_CGraphicContext& gc) const
{	
	FW_CPrivTextBuffer textBuffer(&fString);
	gc.GetRasterizer()->RenderText(gc,
									&textBuffer,
									fPosition,
									fTextAlignment,
									GetRenderVerb(),
									fInk,
									fFont);
}

//----------------------------------------------------------------------------------------
//	FW_CTextShape::RenderText
//----------------------------------------------------------------------------------------

void FW_CTextShape::RenderText(FW_CGraphicContext& gc,
							  	FW_CTextReader* textReader,
							  	const FW_CPoint& position,
							  	const FW_PFont& font,
							 	FW_TextAlignment textAlignment,
							  	const FW_PInk& ink)
{	
	FW_ASSERT(textReader != NULL);
	
	FW_CPrivTextBuffer textBuffer(textReader);
	gc.GetRasterizer()->RenderText(gc,
									&textBuffer,
									position,
									textAlignment,
									FW_kFill,
									ink,
									font);
}

//----------------------------------------------------------------------------------------
//	FW_CTextShape::RenderString
//----------------------------------------------------------------------------------------

void FW_CTextShape::RenderText(FW_CGraphicContext& gc,
							  	const FW_CString& string,
							  	const FW_CPoint& position,
							  	const FW_PFont& font,
							 	FW_TextAlignment textAlignment,
							  	const FW_PInk& ink)
{	
	FW_CPrivTextBuffer textBuffer(&string);
	gc.GetRasterizer()->RenderText(gc,
									&textBuffer,
									position,
									textAlignment,
									FW_kFill,
									ink,
									font);
}

//----------------------------------------------------------------------------------------
//	FW_CTextShape::Transform
//----------------------------------------------------------------------------------------

void FW_CTextShape::Transform(Environment* ev, ODTransform* odTransform)
{
	fPosition.Transform(ev, odTransform);
}

//----------------------------------------------------------------------------------------
//	FW_CTextShape::InverseTransform
//----------------------------------------------------------------------------------------

void FW_CTextShape::InverseTransform(Environment* ev, ODTransform* odTransform)
{
	fPosition.InverseTransform(ev, odTransform);
}

//----------------------------------------------------------------------------------------
//	FW_CTextShape::Inset
//----------------------------------------------------------------------------------------

void FW_CTextShape::Inset(FW_CFixed x, FW_CFixed y)
{
FW_UNUSED(x);
FW_UNUSED(y);

	// Nothing to do
}

//----------------------------------------------------------------------------------------
//	FW_CTextShape::Copy
//----------------------------------------------------------------------------------------

FW_CShape* FW_CTextShape::Copy() const
{
	return FW_NEW(FW_CTextShape, (*this));
}

//----------------------------------------------------------------------------------------
//	FW_CTextShape::MoveShape
//----------------------------------------------------------------------------------------

void FW_CTextShape::MoveShape(FW_CFixed deltaX, FW_CFixed deltaY)
{
	fPosition.x += deltaX;
	fPosition.y += deltaY;
}

//----------------------------------------------------------------------------------------
//	FW_CTextShape::MoveShapeTo
//----------------------------------------------------------------------------------------

void FW_CTextShape::MoveShapeTo(FW_CFixed x, FW_CFixed y)
{
	fPosition.x = x;
	fPosition.y = y;
}

//----------------------------------------------------------------------------------------
//	FW_CTextShape::TextExtent
//----------------------------------------------------------------------------------------

FW_CPoint FW_CTextShape::TextExtent(FW_CGraphicContext& gc) const
{
	FW_CPrivTextBuffer textBuffer(&fString);
	FW_CPoint textExtent;
	gc.GetRasterizer()->TextExtent(gc,
									&textBuffer, 
									fFont,
									textExtent);
	return textExtent;
}

//----------------------------------------------------------------------------------------
//	FW_CTextShape::TextExtent
//----------------------------------------------------------------------------------------

FW_CPoint FW_CTextShape::TextExtent(FW_CGraphicContext& gc,
							  		FW_CTextReader* textReader,
									const FW_PFont& font)
{
	FW_CPrivTextBuffer textBuffer(textReader);
	FW_CPoint textExtent;
	gc.GetRasterizer()->TextExtent(gc,
									&textBuffer, 
									font,
									textExtent);
	return textExtent;
}

//----------------------------------------------------------------------------------------
//	FW_CTextShape::TextExtent
//----------------------------------------------------------------------------------------

FW_CPoint FW_CTextShape::TextExtent(FW_CGraphicContext& gc,
							  		const FW_CString& string,
							  		const FW_PFont& font)
{
	FW_CPrivTextBuffer textBuffer(&string);
	FW_CPoint textExtent;
	gc.GetRasterizer()->TextExtent(gc,
									&textBuffer, 
									font,
									textExtent);
	return textExtent;
}

//----------------------------------------------------------------------------------------
//	FW_CTextShape::GetBounds
//----------------------------------------------------------------------------------------

void FW_CTextShape::GetBounds(FW_CGraphicContext& gc, FW_CRect& rect) const
{
	FW_SPlatformRect plfmRect;
	
	FW_SPlatformPoint position = gc.LogicalToDevice(fPosition);	
	FW_CPrivTextBuffer textBuffer(&fString);
	FW_CPoint textExtent;
	gc.GetRasterizer()->TextExtent(gc, &textBuffer, fFont, textExtent);

	FW_SPlatformPoint plfmExtent = gc.LogicalToDevice(textExtent);
	
	switch (fTextAlignment & FW_kTextAlignPrivHorzAlignMask)
	{
		case FW_kTextAlignLeft:
			plfmRect.left = position.X();
			plfmRect.right = plfmRect.left + plfmExtent.X();
			break;
			
		case FW_kTextAlignRight:
			plfmRect.right = position.X();
			plfmRect.left = plfmRect.right - plfmExtent.X();
			break;
			
		case FW_kTextAlignHCenter:
			plfmRect.left = position.X() - (plfmExtent.X() / 2);
			plfmRect.right = plfmRect.left + plfmExtent.X();
			break;
	}

	FW_SFontMetrics fontMetrics;
	switch (fTextAlignment & FW_kTextAlignPrivVertAlignMask)
	{
		case FW_kTextAlignTop:
			plfmRect.top = position.Y();
			plfmRect.bottom = plfmRect.top + plfmExtent.Y();
			break;
			
		case FW_kTextAlignBottom:
			plfmRect.bottom = position.Y();
			plfmRect.top = plfmRect.bottom - plfmExtent.Y();
			break;
			
		case FW_kTextAlignVCenter:
			fFont->GetFontMetrics(gc.GetGraphicDevice(), fontMetrics);		
			plfmRect.top = position.Y() - (fontMetrics.GetFontHeight() / 2);
			plfmRect.bottom = plfmRect.top + fontMetrics.GetFontHeight();		
			break;
			
		case FW_kTextAlignBaseLine:
			fFont->GetFontMetrics(gc.GetGraphicDevice(), fontMetrics);		
			plfmRect.top = position.Y() - fontMetrics.GetAscent();
			plfmRect.bottom = plfmRect.top + fontMetrics.GetFontHeight();		
			break;
	}
		
	rect = gc.DeviceToLogical(plfmRect);
}

//----------------------------------------------------------------------------------------
//	FW_CTextShape::Flatten
//----------------------------------------------------------------------------------------

void FW_CTextShape::Flatten(FW_CWritableStream& archive) const
{
	FW_CBaseTextShape::Flatten(archive);
	
	archive << fPosition;
	archive << fTextAlignment;
}

//----------------------------------------------------------------------------------------
//	FW_CTextShape::Read
//----------------------------------------------------------------------------------------

void* FW_CTextShape::Read(FW_CReadableStream& archive)
{
	return FW_NEW(FW_CTextShape, (archive));
}

//----------------------------------------------------------------------------------------
//	FW_CTextShape::GetBaseLine
//----------------------------------------------------------------------------------------

void FW_CTextShape::GetBaseLine(FW_CGraphicContext& gc,
								FW_CPoint& start, FW_CPoint& end) const
{	
	FW_SPlatformPoint pt1, pt2;

	FW_SPlatformPoint plfmPos = gc.LogicalToDevice(fPosition);
	
	FW_SFontMetrics fontMetrics;
	fFont->GetFontMetrics(gc.GetGraphicDevice(), fontMetrics);		

	switch (fTextAlignment & FW_kTextAlignPrivVertAlignMask)
	{
		case FW_kTextAlignTop:
			plfmPos.Y() += fontMetrics.GetAscent();
			break;
			
		case FW_kTextAlignBottom:
			plfmPos.Y() -= fontMetrics.GetDescent();
			break;
			
		case FW_kTextAlignVCenter:
			plfmPos.Y() += fontMetrics.GetAscent() - (fontMetrics.GetFontHeight() / 2);
			break;
			
		case FW_kTextAlignBaseLine:
			break;
	}

	FW_CPrivTextBuffer textBuffer(&fString);
	FW_CPoint textExtent;
	gc.GetRasterizer()->TextExtent(gc, &textBuffer, fFont, textExtent);
	FW_SPlatformPoint plfmExtent = gc.LogicalToDevice(textExtent);
	
	switch (fTextAlignment & FW_kTextAlignPrivHorzAlignMask)
	{
		case FW_kTextAlignLeft:
			pt1 = plfmPos;
			pt2.Set(pt1.X() + plfmExtent.X(), pt1.Y());
			break;
			
		case FW_kTextAlignRight:
			pt2 = plfmPos;
			pt1.Set(pt2.X() - plfmExtent.X(), pt2.Y());
			break;
			
		case FW_kTextAlignHCenter:
			pt1.Set(plfmPos.X() - (plfmExtent.X() / 2), plfmPos.Y());
			pt2.Set(pt1.X() + plfmExtent.X(), plfmPos.Y());
			break;
	}

	start = gc.DeviceToLogical(pt1);
	end = gc.DeviceToLogical(pt2);
}

